"""
View utils

"""

from typing import List, Optional

_INDEX_PREFIX = "idx"


def format_offset(offset: str) -> str:
    """Format an offset interval"""
    # @TODO check valid intervals?
    return "INTERVAL '{}'".format(offset)


def get_index_name(table_name: str, keys: Optional[List[str]] = None, unique: bool = False) -> str:
    keys_list = "_".join(keys) if keys else None

    name_components = [
        _INDEX_PREFIX,
        table_name,
        keys_list,
        "unique" if unique else "",
    ]

    index_name = "_".join([i for i in name_components if i])

    return index_name



from typing import Dict, List

from geojson_pydantic.geometries import Point

from opennem.db.models.opennem import Station

from .schema import FacilityFeature, FacilityGeo


def stations_to_geojson(stations: List[Station]) -> FacilityGeo:
    features = []

    for station in stations:
        if not station.location:
            continue

        if not station.facilities or len(station.facilities) < 1:
            continue

        feature_dict: Dict = dict(properties=dict())

        feature_dict["properties"] = {
            "station_id": station.id,
            "station_code": station.code,
            "facility_id": station.code,
            "network": station.facilities[0].network.label,
            "network_country": station.facilities[0].network.country,
            "state": station.location.state,
            "postcode": station.location.postcode,
            "name": station.name,
            "capacity_registered": station.capacity_registered,
            # "capacity_aggregate": station.capacity_aggregate,
            "duid_data": [],
        }

        for facility in station.facilities:

            if not facility.fueltech:
                continue

            feature_dict["properties"]["duid_data"].append(
                {
                    # "oid": facility.oid,
                    # "duid": facility.duid,
                    "fuel_tech": facility.fueltech.code,
                    "fuel_tech_label": facility.fueltech.label,
                    "fuel_tech_renewable": facility.fueltech.renewable,
                    "commissioned_date": facility.registered,
                    "decommissioned_date": facility.deregistered,
                    "status": facility.status.code,
                    "status_label": facility.status.label,
                    "unit_id": facility.unit_id,
                    "unit_number": facility.unit_number,
                    "unit_size": facility.unit_capacity,
                    "unit_alias": facility.unit_alias,
                    # capacities for the unit
                    "capacity_registered": facility.capacity_registered,
                    # "capacity_aggregate": facility.capacity_aggregate,
                    # network specific fields (DUID is one)
                    "network_region": facility.network_region,
                }
            )

        feature = FacilityFeature(**feature_dict)

        if station.location and station.location.geom:
            geom = Point(
                coordinates=(station.location.lng, station.location.lat)
            )

            if not feature.geometry:
                feature.geometry = geom

        features.append(feature)

    crs = {
        "type": "name",
        "properties": {"name": "urn:ogc:def:crs:OGC:1.3:CRS84"},
    }

    geo = FacilityGeo(features=features, crs=crs, name="opennem.facilities")

    return geo



# pylint: disable=no-member
"""
Update fueltechs

Revision ID: fcb509301bda
Revises: 6e45ee9557c2
Create Date: 2021-02-04 10:41:38.122003

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "fcb509301bda"
down_revision = "6e45ee9557c2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("update facility set fueltech_id='battery_discharging' where code IN ('LBBL1')")


def downgrade() -> None:
    pass



import csv
import json
import logging
from datetime import datetime
from functools import reduce
from itertools import groupby
from pathlib import Path
from typing import Optional

from opennem.core.facilitystatus import parse_facility_status
from opennem.core.loader import load_data
from opennem.core.normalizers import station_name_cleaner
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.schema.opennem import StationSchema
from opennem.schema.stations import StationSet

logger = logging.getLogger("opennem.importer.mms")


def parse_mms_date(date_string: str) -> Optional[datetime]:
    """

    `25/10/1998  12:00:00 am` => d

    """
    date_string_components = date_string.strip().split(" ")

    if len(date_string_components) < 2:
        raise Exception("Error parsing date: {}".format(date_string))

    date_part = date_string_components[0]

    dt = None

    try:
        dt = datetime.strptime(date_part, "%Y/%m/%d")
    except ValueError:
        raise Exception("Error parsing date: {}".format(date_string))

    # AEMO sets dates to this value when they mean None
    # if dt.year == 2999:
    # return None

    return dt


def dudetailsummary_grouper(tables):

    if "PARTICIPANT_REGISTRATION_DUDETAILSUMMARY" not in tables:
        raise Exception("No dudetailsummary table")

    records = tables["PARTICIPANT_REGISTRATION_DUDETAILSUMMARY"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "date_start": parse_mms_date(i["START_DATE"]),
            "date_end": parse_mms_date(i["END_DATE"]),
            # "status": "operating",
            "code": i["DUID"],
            "network_region": i["REGIONID"],
            "station_code": i["STATIONID"],
            "participant_id": i["PARTICIPANTID"],
            "dispatch_type": i["DISPATCHTYPE"],
        }
        for i in records
    ]

    now = datetime.now()

    records = [
        {
            **i,
            "status": parse_facility_status("retired")
            if i["date_end"] <= now
            else parse_facility_status("operating"),
        }
        for i in records
    ]

    grouped_records = {}

    # First pass sorts facilities into stations
    for k, v in groupby(records, lambda x: (x["station_code"], x["code"])):
        key = k[0]
        duid = k[1]
        if key not in grouped_records:
            grouped_records[key] = {}
            grouped_records[key]["station_code"] = k[0]
            # grouped_records[key]["participant"] = v[0]["PARTICIPANTID"]
            grouped_records[key]["details"] = {}
            grouped_records[key]["facilities"] = []

        if duid not in grouped_records[key]["details"]:
            grouped_records[key]["details"][duid] = []

        grouped_records[key]["details"][duid] += list(v)

    # Second pass flatten the records and we should get start and end dates
    # and a derived status
    for rec in grouped_records.keys():
        for _, facility_group_records in grouped_records[rec][
            "details"
        ].items():

            # date_end_min = min(
            #     facility_group_records, key=lambda x: x["date_end"]
            # )
            date_end_max = max(
                facility_group_records, key=lambda x: x["date_end"]
            )
            date_start_min = min(
                facility_group_records, key=lambda x: x["date_start"]
            )

            grouped_rec = {
                **date_end_max,
                "date_start": date_start_min["date_start"],
            }

            if grouped_rec["date_end"].year == 2999:
                grouped_rec["date_end"] = None

            grouped_records[rec]["facilities"].append(grouped_rec)

    grouped_records = [
        {"station_code": i, "facilities": v["facilities"]}
        for i, v in grouped_records.items()
    ]

    tables["PARTICIPANT_REGISTRATION_DUDETAILSUMMARY"] = grouped_records

    for record in grouped_records:
        station_code = record["station_code"]

        if station_code not in mms:
            print("dudetailsummary: {} not in mms".format(station_code))
            continue

        # mms[station_code] = {**record, **mms[station_code]}
        mms[station_code]["facilities"] = record["facilities"]

    tables["mms"] = mms

    return tables


def operatingstatus_grouper(tables):

    if "PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS" not in tables:
        raise Exception(
            "No PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS table"
        )

    records = tables["PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "effective_date": parse_mms_date(i["EFFECTIVEDATE"]),
            "station_code": i["STATIONID"],
            "status": i["STATUS"],
        }
        for i in records
    ]

    grouped_records = {}

    for station_code, records in groupby(records, lambda x: x["station_code"]):
        if station_code not in grouped_records:
            grouped_records[station_code] = {}
            grouped_records[station_code]["id"] = station_code
            grouped_records[station_code]["details"] = []

        grouped_records[station_code]["details"] += list(records)

    for station_code, record in grouped_records.items():
        date_max = max(record["details"], key=lambda x: x["effective_date"])
        grouped_records[station_code]["status"] = date_max["status"]
        # grouped_records[station_code]["status"] = date_max["status"]

        if station_code not in mms:
            print("operatingstatus: {} is not in MMS".format(station_code))

        mms[station_code]["status"] = date_max["status"]

    tables["PARTICIPANT_REGISTRATION_STATIONOPERATINGSTATUS"] = grouped_records

    tables["mms"] = mms

    return tables


def stations_grouper(tables):

    if "PARTICIPANT_REGISTRATION_STATION" not in tables:
        raise Exception("No PARTICIPANT_REGISTRATION_STATION table")

    records = tables["PARTICIPANT_REGISTRATION_STATION"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "id": _id,
            "updated_at": parse_mms_date(i["LASTCHANGED"]),
            "name": station_name_cleaner(i["STATIONNAME"]),
            "code": i["STATIONID"],
            "station_code": i["STATIONID"],
            "network_name": i["STATIONNAME"],
            "address1": i["ADDRESS1"],
            "address2": i["ADDRESS2"],
            "locality": i["CITY"],
            "state": i["STATE"],
            "postcode": i["POSTCODE"],
            "facilities": [],
        }
        for _id, i in enumerate(records, start=1000)
    ]

    for record in records:
        station_code = record["station_code"]

        if station_code not in mms:
            mms[station_code] = {}

        mms[station_code] = record

    tables["mms"] = mms

    return tables


def dudetail_grouper(tables):

    if "PARTICIPANT_REGISTRATION_DUDETAIL" not in tables:
        raise Exception("No PARTICIPANT_REGISTRATION_DUDETAIL table")

    records = tables["PARTICIPANT_REGISTRATION_DUDETAIL"]

    mms = tables["mms"] if "mms" in tables else {}

    records = [
        {
            "code": i["DUID"],
            "version": i["VERSIONNO"],
            "capacity_registered": float(i["REGISTEREDCAPACITY"]),
            "capacity_maximum": float(i["MAXCAPACITY"]),
            "dispatch_type": i["DISPATCHTYPE"],
        }
        for _id, i in enumerate(records, start=1000)
    ]

    grouped_records = {}

    for network_code, records in groupby(records, lambda x: x["code"]):
        if network_code not in grouped_records:
            grouped_records[network_code] = {}
            grouped_records[network_code]["details"] = []

        grouped_records[network_code]["details"] += list(records)

    for network_code, record in grouped_records.items():
        version_max = max(record["details"], key=lambda x: x["version"])

        dudetail_record = {
            "capacity_registered": version_max["capacity_registered"],
            "capacity_maximum": version_max["capacity_maximum"],
        }

        facility = None

        for station_code in mms.keys():

            for fac in mms[station_code]["facilities"]:
                if fac["code"] == network_code:

                    # don't need the version field any more
                    fac.pop("version", None)

                    facility = {
                        **fac,
                        **dudetail_record,
                    }
                    fac_index = mms[station_code]["facilities"].index(fac)
                    mms[station_code]["facilities"][fac_index] = facility

        if not facility:
            print("dudetail: couldn't find facility: {}".format(network_code))

    tables["PARTICIPANT_REGISTRATION_DUDETAIL"] = grouped_records

    tables["mms"] = mms

    return tables


def genunits_grouper(tables):
    table_name = "PARTICIPANT_REGISTRATION_GENUNITS"

    if table_name not in tables:
        raise Exception("no genunits table")

    records = tables[table_name]

    records = [
        {
            "network_genset": i["GENSETID"],
            "emissions_factor_co2": float(i["CO2E_EMISSIONS_FACTOR"]),
        }
        for _, i in enumerate(records)
        if i["CO2E_EMISSIONS_FACTOR"]
    ]

    mms = tables["mms"] if "mms" in tables else {}

    # grouped_records = {}
    for record in records:
        facility_code = record["network_genset"]

        facility = None

        for station_code in mms.keys():
            for fac in mms[station_code]["facilities"]:
                if fac["code"] == facility_code:

                    facility = {
                        **fac,
                        **record,
                    }
                    fac_index = mms[station_code]["facilities"].index(fac)
                    # print(
                    #     "{} => {}".format(
                    #         station_code, record["emissions_factor_co2"]
                    #     )
                    # )
                    mms[station_code]["facilities"][fac_index][
                        "emissions_factor_co2"
                    ] = record["emissions_factor_co2"]

        if not facility:
            print("dudetail: couldn't find facility: {}".format(facility_code))

    tables[table_name] = records

    tables["mms"] = mms

    return tables


def load_aemo_csv(item, filename):

    if not item:
        item = {}

    if type(item) is not dict:
        raise Exception(
            "Invalid item type expecting a dict so we can fill it "
        )

    current_item = load_data(filename, True, content_type="latin-1")

    if "content" not in current_item:
        logger.error("No content in item to parse")
        return item

    content = current_item["content"]
    del current_item["content"]

    table_name = None
    table_values = ""
    table_fields = []
    table_records = []

    content_split = content.splitlines()

    datacsv = csv.reader(content_split)

    for row in datacsv:
        if not row or type(row) is not list or len(row) < 1:
            continue

        record_type = row[0]

        if record_type == "C":
            # @TODO csv meta stored in table
            if table_name is not None:
                item[table_name] = table_records

        elif record_type == "I":
            if table_name is not None:
                item[table_name] = table_records

            table_name = "{}_{}".format(row[1], row[2])
            table_fields = row[4:]
            table_records = []

        elif record_type == "D":
            table_values = row[4:]
            record = dict(zip(table_fields, table_values))
            table_records.append(record)

    return item


def load_mms_tables():
    mms_path = Path(__file__).parent.parent / "data" / "mms"

    mms_files = [
        "mms/{}".format(f.name)
        for f in mms_path.iterdir()
        if f.suffix in [".zip"]
    ]

    tables = reduce(load_aemo_csv, mms_files, {})

    return tables


def mms_import():
    tables = load_mms_tables()

    logger.info(
        "Imported {} tables: {}".format(
            len(tables.keys()), ", ".join(list(tables.keys()))
        )
    )

    tables = stations_grouper(tables)
    tables = dudetailsummary_grouper(tables)
    tables = operatingstatus_grouper(tables)
    tables = dudetail_grouper(tables)
    tables = genunits_grouper(tables)

    _mms = tables["mms"]

    mms = StationSet()

    for s in _mms.values():
        mms.add_dict(s)

    return mms


def mms_station_map_from_records(mms):
    """
    Get the station to duid map from MMS and return it
    """

    mms_duid_station_map = {}

    for station_record in mms.as_list():
        for network_code in [i.code for i in station_record.facilities]:
            mms_duid_station_map[network_code] = station_record.code

    return mms_duid_station_map


def mms_export():
    """

    Export MMS records

    @TODO move this to opennem.export and keep modules consistent
    """
    mms = mms_import()

    mms_duid_station_map = mms_station_map_from_records(mms)

    with open("data/mms.json", "w") as fh:
        fh.write(mms.json(indent=4))

    with open("data/mms_duid_station_map.json", "w") as fh:
        json.dump(mms_duid_station_map, fh, indent=4, cls=OpenNEMJSONEncoder)


if __name__ == "__main__":
    mms_export()



"""
OpenNEM Diff Versions
"""

import glob
import json
import logging
import os
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Optional
from urllib.parse import urljoin

from opennem.api.export.map import StatType
from opennem.api.stats.loader import load_statset
from opennem.api.stats.schema import OpennemData, OpennemDataSet
from opennem.core.compat.loader import load_statset_v2
from opennem.core.compat.schema import OpennemDataSetV2, OpennemDataV2
from opennem.core.compat.utils import translate_id_v2_to_v3
from opennem.db import SessionLocal
from opennem.db.models.opennem import NetworkRegion
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.schema.core import BaseConfig
from opennem.schema.network import NetworkNEM, NetworkSchema
from opennem.utils.http import http
from opennem.utils.series import series_are_equal, series_not_close

logger = logging.getLogger("opennem.diff.versions")

BASE_URL_V2 = "https://data.opennem.org.au"
BASE_URL_V3 = "https://data.dev.opennem.org.au"

CUR_YEAR = datetime.now().year


def get_v2_url(
    stat_type: StatType, network_region: str, bucket_size: str = "daily", year: int = CUR_YEAR
) -> str:
    """
    get v2 url

    https://data.opennem.org.au/power/tas1.json
    https://data.opennem.org.au/tas1/energy/daily/2021.json
    https://data.opennem.org.au/tas1/energy/monthly/all.json
    """

    url_components = []

    if stat_type == StatType.power:
        url_components = ["power", network_region.lower()]

    elif stat_type == StatType.energy and bucket_size == "monthly":
        url_components = [network_region.lower(), "energy", bucket_size, "all"]

    elif stat_type == StatType.energy and bucket_size == "daily":
        url_components = [network_region.lower(), "energy", bucket_size, str(year)]

    else:
        raise Exception("Invalid v2 url components")

    url_path = "/".join(url_components)
    url_path += ".json"

    url = urljoin(BASE_URL_V2, url_path)

    return url


def get_v3_url(
    stat_type: StatType, network_region: str, bucket_size: str = "daily", year: int = CUR_YEAR
) -> str:
    """
    Get v3 url

    https://data.dev.opennem.org.au/v3/stats/au/NEM/TAS1/power/7d.json
    https://data.dev.opennem.org.au/v3/stats/au/NEM/TAS1/energy/2021.json
    https://data.dev.opennem.org.au/v3/stats/au/NEM/TAS1/energy/all.json
    """

    url_components = [
        "v3",
        "stats",
        "au",
        "NEM",
        network_region.upper(),
        stat_type.value,
    ]

    if stat_type == StatType.power:
        url_components += [
            "7d",
        ]

    elif stat_type == StatType.energy and bucket_size == "monthly":
        url_components += ["all"]

    elif stat_type == StatType.energy and bucket_size == "daily":
        url_components += [str(year)]

    else:
        raise Exception("Invalid v2 url components")

    url_path = "/".join(url_components)
    url_path += ".json"

    url = urljoin(BASE_URL_V3, url_path)

    return url


class DiffComparisonSet(BaseConfig):
    stat_type: StatType
    network_region: str
    bucket_size: Optional[str]
    year: Optional[int]

    v2: Optional[OpennemDataSetV2]
    v3: Optional[OpennemDataSet]

    @property
    def urlv2(self) -> str:
        return get_v2_url(self.stat_type, self.network_region, self.bucket_size, self.year)

    @property
    def urlv3(self) -> str:
        return get_v3_url(self.stat_type, self.network_region, self.bucket_size, self.year)

    def load_maps(self) -> None:
        for version in ["v2", "v3"]:
            req_url = getattr(self, f"url{version}")

            r = http.get(req_url)
            logger.debug("Loading: {}".format(req_url))

            if not r.ok:
                logger.error("invalid {} url: {}".format(version, req_url))

            statset = None

            if version == "v2":
                statset = load_statset_v2(r.json())
                self.v2 = statset

            else:
                statset = load_statset(r.json())
                self.v3 = statset


def get_url_map(regions: List[NetworkRegion]) -> List[DiffComparisonSet]:
    urls = []

    for region in regions:
        a = DiffComparisonSet(
            stat_type=StatType.power, network_region=region.code, bucket_size="7d"
        )
        urls.append(a)

        for y in range(2017, 2022):
            a = DiffComparisonSet(
                stat_type=StatType.energy, network_region=region.code, bucket_size="daily", year=y
            )
            urls.append(a)

        a = DiffComparisonSet(
            stat_type=StatType.energy, network_region=region.code, bucket_size="monthly"
        )
        urls.append(a)

    return urls


def validate_url_map(url_map: List[DiffComparisonSet]) -> bool:
    success = True

    for us in url_map:
        for version in ["v2", "v3"]:
            req_url = getattr(us, f"url{version}")
            r = http.get(req_url)

            if not r.ok:
                logger.error("invalid {} url: {}".format(version, req_url))
                success = False

    return success


def load_url_map(url_map: List[DiffComparisonSet]) -> List[DiffComparisonSet]:

    for us in url_map:
        for version in ["v2", "v3"]:
            req_url = getattr(us, f"url{version}")
            r = http.get(req_url)

            if not r.ok:
                logger.error("invalid {} url: {}".format(version, req_url))

            statset = None

            if version == "v2":
                statset = load_statset_v2(r.json())
                us.v2 = statset

            else:
                statset = load_statset(r.json())
                us.v3 = statset

    return url_map


def get_network_regions(
    network: NetworkSchema, network_region: Optional[str] = None
) -> List[NetworkRegion]:
    """Return regions for a network"""
    s = SessionLocal()
    regions = s.query(NetworkRegion).filter_by(network_id=network.code)

    if network_region:
        regions = regions.filter_by(code=network_region)

    regions = regions.all()

    return regions


def get_id_diff(seriesv2: OpennemDataV2, seriesv3: OpennemData) -> Optional[List[str]]:
    id2_diffs = sorted([i.id for i in seriesv2])
    id3_diffs = sorted([i.id_v2() for i in seriesv3])

    diff = list(set(id2_diffs) - set(id3_diffs))

    if len(diff) < 1:
        return None

    return diff


def get_data_by_id(id: str, series: List[Dict]) -> Optional[Dict]:
    id_s = list(filter(lambda x: x["id"] == id, series))

    if len(id_s) < 1:
        logger.error("Could not find id {} in series".format(id))
        return None

    return id_s.pop()


def run_diff() -> float:
    score = 0
    score_tested = 0
    buckets_not_match = 0

    buckets_total = 0
    buckets_matches = 0

    regions = get_network_regions(NetworkNEM, "NSW1")
    statsetmap = get_url_map(regions)

    # load urls
    [i.load_maps() for i in statsetmap]

    # filter it down for now
    # statsetmap_power = list(filter(lambda x: x.stat_type == StatType.power, statsetmap))

    for statset in statsetmap:
        if not statset.v2 or not statset.v3:
            logger.error("Error getting schemas for v2 or v3")
            continue

        logger.info(
            "Comparing {} {} for {}".format(
                statset.stat_type.value, statset.bucket_size, statset.network_region
            )
        )
        logger.info(statset.urlv2)
        logger.info(statset.urlv3)

        if len(statset.v2.data) != len(statset.v2.data):
            logger.info(
                "Series {} {} for {} is missing ids. v2 has {} v3 has {}".format(
                    statset.stat_type.value,
                    statset.bucket_size,
                    statset.network_region,
                    len(statset.v2.data),
                    len(statset.v3.data),
                )
            )

        id2_diffs = sorted([i.id for i in statset.v2.data])
        id3_diffs = sorted([i.id_v2() for i in statset.v3.data])

        logger.info("v2 has ids:")
        for i in id2_diffs:
            logger.info("\t{}".format(i))

        logger.info("v3 has ids:")
        for i in id3_diffs:
            logger.info("\t{}".format(i))

        id_diff = list(set(id2_diffs) - set(id3_diffs))

        if len(id_diff) > 0:
            for i in id_diff:
                logger.error(
                    "Series {} {} for {} is missing ids. v3 doesn't have {}".format(
                        statset.stat_type.value, statset.bucket_size, statset.network_region, i
                    )
                )

        logger.info("Comparing each id field")

        for i in id2_diffs:
            logger.info(" = comparing {}".format(i))

            v2i = statset.v2.get_id(i)
            v3i = statset.v3.get_id(translate_id_v2_to_v3(i))

            if not v3i:
                logger.error("    missing in v3: {}".format(translate_id_v2_to_v3(i)))
                continue

            if not v2i:
                logger.error("    error getting id {} from v2".format(i))
                continue

            if v2i.fuel_tech:
                if v2i.fuel_tech == v3i.fueltech_v2():
                    logger.info("  * fueltech matches")
                else:
                    logger.error(
                        "  * fueltech DOESNT MATCH: {} and {}".format(
                            v2i.fuel_tech, v3i.fueltech_v2()
                        )
                    )

            if "temperature" in v2i.id:
                continue

            if "emissions" in v2i.id:
                continue

            if v2i.history:
                logger.info("  * comparing history:")
                score_tested += 1

                if len(v2i.history.data) != len(v3i.history.data):
                    logger.error(
                        "    - data length mismatch v2 {} v3 {}".format(
                            len(v2i.history.data), len(v3i.history.data)
                        )
                    )

                data_matches = series_are_equal(
                    v2i.history.values(), v3i.history.values(), full_equality=True
                )

                buckets_total += len(data_matches.keys())

                if False in list(data_matches.values()):
                    logger.error("    - values don't match ")

                    mismatch_values = series_not_close(
                        v2i.history.values(), v3i.history.values(), full_equality=True
                    )

                    score += 1
                    buckets_not_match += len(mismatch_values.keys())

                    extra_part = i.split(".")[1]
                    units = i.split(".")[-1]

                    file_components = [
                        units,
                        statset.network_region,
                        statset.bucket_size,
                        extra_part,
                        v3i.fueltech_v2(),
                        statset.year,
                    ]

                    filename = "-".join([str(i) for i in file_components if i])

                    with open(
                        f"../dataquality/diff/{filename}-diff.json",
                        "w",
                    ) as fh:
                        json.dump(mismatch_values, fh, cls=OpenNEMJSONEncoder, indent=4)

                elif data_matches.keys() and len(data_matches.keys()):
                    logger.info(
                        "     - series values match {} values between {} and {}".format(
                            len(data_matches.keys()),
                            min(data_matches.keys()),
                            max(data_matches.keys()),
                        )
                    )
                else:
                    logger.error(
                        "    - Error matching data values: {}".format(len(data_matches.keys()))
                    )

            # remaining attributes
            for v2ikey in v2i.dict().keys():
                if v2ikey in ["id", "fuel_tech", "history", "forecast"]:
                    continue

                if not hasattr(v3i, v2ikey):
                    logger.error("  * key {} is missing".format(v2ikey))
                else:
                    data_matches = getattr(v2i, v2ikey) == getattr(v3i, v2ikey)

                    if data_matches:
                        logger.info("  * key {} exists and matches ".format(v2ikey))
                    else:
                        logger.error(
                            "  * key {} DOESNT MATCH. values: v2'{}' and v3'{}'".format(
                                v2ikey, getattr(v2i, v2ikey), getattr(v3i, v2ikey)
                            )
                        )

        logger.info("=" * 50)

    if score == 0:
        return 0.0

    if buckets_not_match == 0:
        percentage = 0.0
    else:
        percentage = round((buckets_total - buckets_not_match) / buckets_total * 100, 2)

    return "{}% match. {} total buckets match of {}.".format(
        percentage, (buckets_total - buckets_not_match), buckets_total
    )


def commit_diffs(score: str) -> None:
    from git import Actor, Repo

    rw_dir = Path(__file__).parent.parent.parent.parent / "dataquality"

    if not rw_dir.is_dir():
        raise Exception("not a git directory: {}".format(rw_dir))

    repo = Repo(rw_dir)
    author = committer = Actor("OpenNEM Data Quality", "sysadmin@opennem.org.au")

    repo.index.add(["diff"])
    repo.index.commit(
        "Commit version diffs. Score: {}".format(score), author=author, committer=committer
    )

    origin = repo.remote(name="origin")
    origin.push()


def run_data_diff() -> None:
    regions = get_network_regions(NetworkNEM, "NSW1")
    statsetmap = get_url_map(regions)

    # validate all urls are valid
    validate_url_map(statsetmap)
    statsetmap = load_url_map(statsetmap)

    for statset in statsetmap:
        print("{} {} {} ".format(statset.stat_type, statset.bucket_size, statset.network_region))
        print(statset.v2)
        print(statset.v3)


if __name__ == "__main__":

    files = glob.glob("../dataquality/diff/*")

    for f in files:
        if f in ["empty"]:
            continue

        os.remove(f)
        logger.debug(f"Deleted {f}")

    score = run_diff()

    print(score)
    # commit_diffs(score)



# pylint: disable=no-member
"""
Total demand on balancing summary per region

Revision ID: d130be833a0c
Revises: 1057033fe1ea
Create Date: 2021-01-21 16:54:01.264342

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "d130be833a0c"
down_revision = "1057033fe1ea"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.add_column(
        "balancing_summary",
        sa.Column("demand_total", sa.Numeric(), nullable=True),
    )


def downgrade() -> None:
    op.drop_column("balancing_summary", "demand_total")



# pylint: disable=no-member
"""Make network_region in balancing summary primary

Revision ID: a76e6ff50cda
Revises: 911efdc0c77f
Create Date: 2020-10-07 11:37:59.674373

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "a76e6ff50cda"
down_revision = "911efdc0c77f"
branch_labels = None
depends_on = None


def upgrade():
    op.execute(
        "alter table balancing_summary drop constraint balancing_summary_pkey"
    )
    op.execute(
        "alter table balancing_summary add constraint balancing_summary_pkey primary key (trading_interval, network_id, network_region)"
    )


def downgrade():
    op.execute(
        "alter table balancing_summary drop constraint balancing_summary_pkey"
    )
    op.execute(
        "alter table balancing_summary add constraint balancing_summary_pkey primary key (trading_interval, network_id)"
    )



import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.db import SessionLocal, get_database_engine, get_database_session
from opennem.db.models.opennem import Facility, FacilityScada
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)


class NPIStoreFacility(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if "records" not in item:
            logger.error("Invalid return response")

        records = item["records"]

        engine = get_database_engine()
        session = SessionLocal()

        records_to_store = []

        for record in records:
            records_to_store.append(
                {
                    # "network_id": "WEM" if state == "WA" else "NEM",
                    # "trading_interval": interval_time,
                    # "facility_code": facility_code,
                    # "generated": generated,
                }
            )

        # free
        primary_keys = None

        stmt = insert(FacilityScada).values(records_to_store)
        stmt.bind = engine
        stmt = stmt.on_conflict_do_update(
            index_elements=["trading_interval", "network_id", "facility_code", "is_forecast"],
            set_={
                "generated": stmt.excluded.generated,
            },
        )

        try:
            session.execute(stmt)
            session.commit()
        except Exception as e:
            logger.error("Error: {}".format(e))
        finally:
            session.close()

        return len(records_to_store)



# pylint: disable=no-member
"""
Update indexes on mv_facility_all

Revision ID: 977f219c8563
Revises: f21b1048faf6
Create Date: 2021-01-19 15:24:52.893563

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "977f219c8563"
down_revision = "f21b1048faf6"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("drop index if exists idx_mv_facility_all_trading_interval_day_aest;")

    op.execute(
        "CREATE UNIQUE INDEX IF NOT EXISTS idx_mv_facility_unique ON mv_facility_all (trading_interval, network_id, code);"
    )


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
Create stats hypertable

Revision ID: 43f1d539490f
Revises: 889a1453be90
Create Date: 2021-01-05 17:27:45.071634

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "43f1d539490f"
down_revision = "889a1453be90"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "select create_hypertable('stats', 'stat_date', migrate_data=>true, if_not_exists=>true)"
    )


def downgrade() -> None:
    pass



import logging
from typing import Optional

from opennem.db import get_database_engine
from opennem.utils.dates import subtract_days

logger = logging.getLogger("opennem.db.tasks")

TIMESCALE_VIEWS = [
    "mv_facility_energy_hour",
    "mv_interchange_energy_nem_region",
]

MATERIAL_VIEWS = [
    # "mv_facility_all",
    "mv_facility_30min_all",
    "mv_region_emissions",
    "mv_interchange_energy_nem_region",
]


def refresh_timescale_views(
    view_name: Optional[str] = None, all: bool = False, days: int = 7
) -> None:
    """refresh timescale views"""
    __query = """
    CALL refresh_continuous_aggregate(
        continuous_aggregate => '{view}',
        window_start         => {date_from},
        window_end           => NULL
    );
    """

    engine = get_database_engine()
    dt = subtract_days(days=days)
    date_from = "'{}'".format(dt.strftime("%Y-%m-%d"))

    if all:
        date_from = "NULL"

    views = []

    if view_name:
        views.append(view_name)
    else:
        views = TIMESCALE_VIEWS

    with engine.connect() as c:
        for v in TIMESCALE_VIEWS:
            query = __query.format(view=v, date_from=date_from)
            logger.debug(query)

            try:
                c.execution_options(isolation_level="AUTOCOMMIT").execute(query)
            except Exception as e:
                logger.error("Could not run refresh: {}".format(e))


def refresh_material_views(view_name: Optional[str] = None) -> None:
    """Refresh material views"""
    __query = "REFRESH MATERIALIZED VIEW CONCURRENTLY {view} with data"

    engine = get_database_engine()

    views = []

    if view_name:
        views.append(view_name)
    else:
        views = MATERIAL_VIEWS

    with engine.connect() as c:
        for v in views:
            query = __query.format(view=v)
            logger.debug(query)

            try:
                c.execution_options(isolation_level="AUTOCOMMIT").execute(query)
            except Exception as e:
                logger.error("Could not run material refresh: {}".format(e))


def refresh_views() -> None:
    refresh_timescale_views(all=True)
    refresh_material_views()


if __name__ == "__main__":
    refresh_views()



# pylint: disable=no-member
"""
Update mv_energy_facility_hour mv to include network_id

Revision ID: 2f198fa45c76
Revises: fbd4ae450de2
Create Date: 2021-01-14 01:58:05.502904

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "2f198fa45c76"
down_revision = "fbd4ae450de2"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
    drop view if exists mv_facility_energy_hour cascade;

    CREATE OR REPLACE VIEW mv_facility_energy_hour WITH (timescaledb.continuous) AS
    select
        time_bucket('1 hour', fs.trading_interval) as trading_interval,
        fs.facility_code,
        fs.network_id,
        energy_sum(fs.generated, '1 hour') * interval_size('1 hour', count(fs.generated)) as energy
    from facility_scada fs
    group by
        1, 2, 3;

    """
    )


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""export_set field in network and network region

Revision ID: 0716d393ff9d
Revises: 0bd670d87da9
Create Date: 2020-11-17 02:07:16.413677

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "0716d393ff9d"
down_revision = "0bd670d87da9"
branch_labels = None
depends_on = None


def upgrade():
    # network table
    op.add_column(
        "network", sa.Column("export_set", sa.Boolean(), nullable=True)
    )
    op.execute("UPDATE network set export_set=True")
    op.alter_column("network", "export_set", nullable=False)

    # network region table
    op.add_column(
        "network_region", sa.Column("export_set", sa.Boolean(), nullable=True)
    )
    op.execute("UPDATE network_region set export_set=True")
    op.alter_column("network_region", "export_set", nullable=False)


def downgrade():
    op.drop_column("network_region", "export_set")
    op.drop_column("network", "export_set")



from textwrap import dedent
from typing import Optional

from .schema import ViewDefinition


def create_continuous_aggregation_query(viewdef: ViewDefinition) -> Optional[str]:
    """Get an update query for a continuous aggregation policy"""

    if not viewdef.aggregation_policy:
        return None

    __query = """
    SELECT add_continuous_aggregate_policy(
        '{view_name}',
        start_offset => {start_offset},
        end_offset => {end_offset},
        schedule_interval => {schedule_interval}
    );

    """

    query = __query.format(
        view_name=viewdef.name,
        start_offset=viewdef.aggregation_policy.start_offset,
        end_offset=viewdef.aggregation_policy.end_offset,
        schedule_interval=viewdef.aggregation_policy.schedule_interval,
    )

    return dedent(query)


def remove_continuous_aggregation_query(viewdef: ViewDefinition) -> Optional[str]:
    """ Remove a continuous aggregation policy """

    if not viewdef.aggregation_policy:
        return None

    __query = "SELECT remove_continuous_aggregate_policy('{view_name}');"

    query = __query.format(view_name=viewdef.name)

    return dedent(query)



import logging
from concurrent.futures import ThreadPoolExecutor
from datetime import datetime, timedelta
from typing import Dict, List

from pytz import FixedOffset

from opennem.api.stats.controllers import get_scada_range
from opennem.api.time import human_to_interval, human_to_period
from opennem.core.compat.energy import energy_sum_compat
from opennem.db import get_database_engine
from opennem.db.models.opennem import FacilityScada
from opennem.db.tasks import refresh_timescale_views, refresh_views
from opennem.notifications.slack import slack_message
from opennem.pipelines.bulk_insert import build_insert_query
from opennem.pipelines.csv import generate_csv_from_records
from opennem.schema.dates import DatetimeRange, TimeSeries
from opennem.schema.network import NetworkNEM

logger = logging.getLogger("opennem.workers.energy")


def get_generated(network_region: str, date_min: datetime, date_max: datetime) -> List[Dict]:
    __sql = """
    select
        fs.trading_interval at time zone 'AEST' as trading_interval,
        fs.facility_code,
        fs.network_id,
        generated
    from
        facility_scada fs
        left join facility f on fs.facility_code = f.code
    where
        fs.network_id='NEM'
        and f.network_region='{network_region}'
        and fs.trading_interval >= '{date_min}'
        and fs.trading_interval <= '{date_max}'
        and fs.is_forecast is False
    order by fs.trading_interval asc;
    """

    query = __sql.format(network_region=network_region, date_min=date_min, date_max=date_max)

    engine = get_database_engine()

    results = []

    with engine.connect() as c:
        logger.debug(query)
        results = list(c.execute(query))

    logger.debug("Got back {} rows".format(len(results)))

    return results


def insert_energies(results: List[Dict]) -> None:
    # Get the energy sums
    energy_sum = energy_sum_compat(results)

    # Add metadata
    energy_sum["created_by"] = "opennem.worker.energy"
    energy_sum["created_at"] = ""
    energy_sum["updated_at"] = datetime.now()
    energy_sum["generated"] = None
    energy_sum["is_forecast"] = False

    # reorder columns
    columns = [
        "created_by",
        "created_at",
        "updated_at",
        "network_id",
        "trading_interval",
        "facility_code",
        "generated",
        "eoi_quantity",
        "is_forecast",
    ]
    energy_sum = energy_sum[columns]

    records_to_store: List[Dict] = energy_sum.to_dict("records")

    # Build SQL + CSV and bulk-insert
    sql_query = build_insert_query(FacilityScada, ["updated_at", "eoi_quantity"])
    conn = get_database_engine().raw_connection()
    cursor = conn.cursor()
    csv_content = generate_csv_from_records(
        FacilityScada,
        records_to_store,
        column_names=records_to_store[0].keys(),
    )

    cursor.copy_expert(sql_query, csv_content)
    conn.commit()

    logger.info("Inserted {} records".format(len(records_to_store)))


def get_date_range() -> DatetimeRange:
    date_range = get_scada_range(network=NetworkNEM)

    time_series = TimeSeries(
        start=date_range.start,
        end=date_range.end,
        interval=human_to_interval("1d"),
        period=human_to_period("all"),
        network=NetworkNEM,
    )

    return time_series.get_range()


def run_energy_calc(region: str, date_min: datetime, date_max: datetime) -> str:
    results = get_generated(region, date_min, date_max)

    try:
        insert_energies(results)
        pass
    except Exception as e:
        logger.error(e)
        slack_message("Energy archive error: {}".format(e))

    return "Done {} for {} => {}".format(region, date_min, date_max)


def worker_done(future) -> None:
    logger.debug(future.result())


def run_energy_update_archive() -> None:
    date_range = get_date_range()

    executor = ThreadPoolExecutor(max_workers=4)

    for year in [2020, 2021, 2019, 2018, 2017, 2016]:
        for month in range(1, 12):
            date_min = datetime(
                year=year, month=month, day=1, hour=0, minute=0, second=0, tzinfo=FixedOffset(600)
            ) - timedelta(minutes=5)
            date_max = datetime(
                year=year,
                month=month + 1,
                day=1,
                hour=0,
                minute=5,
                second=0,
                tzinfo=FixedOffset(600),
            )

            if date_max > date_range.end:
                date_max = date_range.end

            if date_min > date_max:
                slack_message("Reached end of energy archive")
                logger.debug("reached end of archive")
                break

            for region in ["QLD1", "NSW1", "VIC1", "TAS1", "SA1"]:
                future = executor.submit(run_energy_calc, region, date_min, date_max)
                future.add_done_callback(worker_done)

    slack_message("Updated energies for {}".format(year))


def run_energy_update() -> None:
    for region in ["QLD1", "NSW1", "VIC1", "TAS1", "SA1"]:
        now = datetime.now().replace(tzinfo=FixedOffset(600))
        date_min = now.replace(hour=0, minute=0, second=0)
        date_max = date_min + timedelta(days=1, minutes=5)

        results = get_generated(region, date_min, date_max)

        try:
            insert_energies(results)
        except Exception as e:
            logger.error(e)
            slack_message("Energy archive error: {}".format(e))


if __name__ == "__main__":
    run_energy_update_archive()
    # refresh_timescale_views()
    # refresh_views()



"""
OpenNEM v2 compatible version of energy sums

Uses an average in 30 minute buckets
"""
from datetime import date, datetime, timedelta
from typing import Dict, Generator, List, Optional, Tuple, Union

import pandas as pd
from pytz import FixedOffset

from opennem.schema.core import BaseConfig


class ScadaResultCompat(BaseConfig):
    interval: datetime
    code: Optional[str]
    generated: Union[float, int, None]


def __trading_energy_generator(
    df: pd.DataFrame, duid: str, sel_date: datetime
) -> Generator[Tuple[date, str, str, float], None, None]:
    df.sort_index(inplace=True)
    t_start = datetime(sel_date.year, sel_date.month, sel_date.day, 0, 5)

    # 48 trading intervals in the day
    # (could be better with groupby function)
    for TI in range(48):
        # t_i initial timestamp of trading_interval, t_f = final timestamp of trading interval
        t_i = t_start + timedelta(0, 1800 * TI)
        t_f = t_start + timedelta(0, 1800 * (TI + 1))

        d_ti = df[(df.index >= t_i) & (df.index <= t_f) & (df.facility_code == duid)]

        if d_ti.empty or d_ti.size < 7:
            continue

        yield d_ti.index[-2].replace(tzinfo=FixedOffset(600)), duid, d_ti.network_id.unique()[
            0
        ], __trapezium_integration(d_ti.generated)


def __trapezium_integration(d_ti: pd.Series) -> float:
    if d_ti.count() < 7:
        return 0

    weights = d_ti.values * [1, 2, 2, 2, 2, 2, 1]

    weights_sum = weights.sum()

    bucket_energy = 0.5 * weights_sum / 12

    return bucket_energy


def energy_sum_compat(gen_series: List[Dict]) -> pd.DataFrame:
    df = pd.DataFrame(
        gen_series,
        columns=[
            "trading_interval",
            "facility_code",
            "network_id",
            "generated",
        ],
    )
    df.trading_interval = pd.to_datetime(df.trading_interval)
    df.generated = pd.to_numeric(df.generated)
    df = df.set_index(df.trading_interval)

    sel_date = df.index.min().date()
    max_date = df.index.max().date()

    return_list = []

    for duid in list(df.facility_code.unique()):
        sel_date = df.trading_interval.min().date()

        while sel_date <= max_date:
            next_frame = [d for d in __trading_energy_generator(df, duid, sel_date)]
            return_list += next_frame
            sel_date = sel_date + timedelta(days=1)

    return_frame = pd.DataFrame(
        return_list,
        columns=[
            "trading_interval",
            "facility_code",
            "network_id",
            "eoi_quantity",
        ],
    )

    return_frame = return_frame.set_index(return_frame.trading_interval)

    return return_frame



"""
OpenNEM API Client

(c) 2021 - Opennem Inc.
"""

from typing import Optional

from opennem.api.stats.schema import (
    DataQueryResult,
    OpennemDataSet,
    RegionFlowResult,
    ScadaDateRange,
)
from opennem.schema.network import NetworkSchema
from opennem.settings.schema import OpennemSettings
from opennem.utils.http import http


class OpennemClient:

    _http = None

    def __init__(self, settings: Optional[OpennemSettings] = None) -> None:
        self._http = http
        pass

    def get_networks() -> None:
        pass


def main() -> None:
    co = OpennemDataSet()

    co.append_set()

    return None


if __name__ == "__main__":
    pass



import logging

from sqlalchemy.dialects.postgresql import insert

from opennem.core.normalizers import clean_float
from opennem.db import SessionLocal, get_database_engine
from opennem.db.models.opennem import BalancingSummary
from opennem.schema.network import NetworkNEM
from opennem.utils.dates import parse_date
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger("opennem.pipeline.nem.summary")


class NemwebSummaryPipeline(object):
    @check_spider_pipeline
    def process_item(self, item, spider=None):
        if not item:
            logger.error("No item in pipeline")
            return None

        if "content" not in item:
            logger.error("No content in pipeline")
            return None

        content = item["content"]

        if "ELEC_NEM_SUMMARY" not in content:
            logger.error("Invalid summary return")
            return None

        summary = content["ELEC_NEM_SUMMARY"]

        s = SessionLocal()
        records_to_store = []

        for row in summary:
            trading_interval = parse_date(row["SETTLEMENTDATE"], network=NetworkNEM)

            demand_total = clean_float(row["TOTALDEMAND"])
            generation_scheduled = clean_float(row["SCHEDULEDGENERATION"])
            generation_non_scheduled = clean_float(row["SEMISCHEDULEDGENERATION"])

            records_to_store.append(
                {
                    "created_by": spider.name,
                    "trading_interval": trading_interval,
                    "network_id": "NEM",
                    "network_region": row["REGIONID"].strip(),
                    # "demand_total": demand_total,
                    "generation_scheduled": generation_scheduled,
                    "generation_non_scheduled": generation_non_scheduled,
                }
            )

        stmt = insert(BalancingSummary).values(records_to_store)
        stmt.bind = get_database_engine()
        stmt = stmt.on_conflict_do_update(
            index_elements=[
                "trading_interval",
                "network_id",
                "network_region",
            ],
            set_={
                # "demand_total": stmt.excluded.demand_total,
                "generation_scheduled": stmt.excluded.generation_scheduled,
                "generation_non_scheduled": stmt.excluded.generation_non_scheduled,
            },
        )

        try:
            s.execute(stmt)
            s.commit()
        except Exception as e:
            logger.error("Error inserting records")
            logger.error(e)
            return {"num_records": 0}
        finally:
            s.close()

        return {"num_records": len(records_to_store)}



import re

from opennem.pipelines.bulk_insert import BulkInsertPipeline
from opennem.pipelines.csv import RecordsToCSVPipeline
from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebLatestRooftopForecast(NemwebSpider):
    name = "au.nem.latest.rooftop_forecast"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/ROOFTOP_PV/FORECAST/"
    limit = 2

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebCurrentRooftopForecast(NemwebSpider):
    name = "au.nem.current.rooftop_forecast"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/ROOFTOP_PV/FORECAST/"

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebArchiveRooftopForecast(NemwebSpider):
    name = "au.nem.archive.rooftop_forecast"
    start_url = "http://www.nemweb.com.au/Reports/ARCHIVE/ROOFTOP_PV/FORECAST/"

    filename_filter = re.compile(r".*_MEASUREMENT_.*")

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )



# pylint: disable=no-member
"""
Update mv_facility_all

Revision ID: 4503ef797f3a
Revises: 2423c1708beb
Create Date: 2021-01-28 00:59:26.822175

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "4503ef797f3a"
down_revision = "2423c1708beb"
branch_labels = None
depends_on = None


stmt_drop = "drop materialized view if exists mv_facility_all cascade;"

stmt = """
create materialized view mv_facility_all as
select
    fs.trading_interval,
    f.code,
    f.fueltech_id,
    f.network_id,
    f.network_region,
    f.interconnector,
    f.interconnector_region_to,
    date_trunc('day', fs.trading_interval at time zone 'AEST') as ti_day_aest,
    date_trunc('month', fs.trading_interval at time zone 'AEST') as ti_month_aest,
    date_trunc('day', fs.trading_interval at time zone 'AWST') as ti_day_awst,
    date_trunc('month', fs.trading_interval at time zone 'AWST') as ti_month_awst,
    max(fs.energy) as energy,
    case when avg(bs.price) >= 0  and min(fs.energy) >= 0 then
        coalesce(
            max(fs.energy) * avg(bs.price),
            0.0
        )
    else 0.0
    end as market_value,
    case when avg(f.emissions_factor_co2) >= 0  and min(fs.energy) >= 0 then
        coalesce(
            max(fs.energy) * avg(f.emissions_factor_co2),
            0.0
        )
    else 0.0
    end as emissions
from mv_facility_energy_hour fs
    left join facility f on fs.facility_code = f.code
    left join balancing_summary bs on
        bs.trading_interval = fs.trading_interval
        and bs.network_id=f.network_id
        and bs.network_region = f.network_region
where
    f.fueltech_id is not null
group by
    1,
    f.code,
    f.fueltech_id,
    f.network_id,
    f.network_region,
    f.interconnector,
    f.interconnector_region_to
order by 1 desc;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



import logging
from operator import attrgetter
from pathlib import Path

from opennem.db import get_database_engine
from opennem.db.views.queries import (
    get_all_views_query,
    get_query_drop_view,
    get_view_unique_index_query,
)

from .continuous_aggregates import (
    create_continuous_aggregation_query,
    remove_continuous_aggregation_query,
)
from .schema import ContinuousAggregationPolicy, ViewDefinition

logger = logging.getLogger("opennem.db.views")

VIEW_PATH = Path(__file__).parent.parent / "fixtures" / "views"


AggregationPolicy30Minutes = ContinuousAggregationPolicy(
    interval="30 minutes", start_interval="2 hours"
)

AggregationPolicy2Hours = ContinuousAggregationPolicy(
    interval="2 hours", start_interval="6 hours", end_interval="2 hours"
)

AggregationPolicy6Hours = ContinuousAggregationPolicy(
    interval="6 hours", start_interval="12 hours", end_interval="2 hours"
)

_VIEW_MAP = [
    # ViewDefinition(
    #     priority=1,
    #     name="mv_facility_energy_hour",
    #     materialized=True,
    #     filepath="mv_facility_energy_hour.sql",
    #     aggregation_policy=AggregationPolicy2Hours,
    # ),
    ViewDefinition(
        priority=2,
        name="mv_facility_energy_30m",
        materialized=True,
        filepath="mv_facility_energy_30min.sql",
        aggregation_policy=AggregationPolicy2Hours,
    ),
    # ViewDefinition(
    #     priority=10,
    #     name="mv_facility_all",
    #     materialized=True,
    #     filepath="mv_facility_all.sql",
    #     primary_key=["trading_interval", "network_id", "code"],
    # ),
    ViewDefinition(
        priority=11,
        name="mv_facility_30min_all",
        materialized=True,
        filepath="mv_facility_30min_all.sql",
        primary_key=["trading_interval", "network_id", "code"],
    ),
    ViewDefinition(
        priority=20,
        name="mv_region_emissions",
        materialized=True,
        filepath="mv_region_emissions.sql",
        primary_key=["trading_interval", "network_id", "network_region"],
    ),
    ViewDefinition(
        priority=30,
        name="mv_interchange_energy_nem_region",
        materialized=True,
        filepath="mv_interchange_energy_nem_region.sql",
    ),
    ViewDefinition(
        priority=40,
        name="vw_region_flow_emissions",
        materialized=False,
        filepath="vw_region_flow_emissions.sql",
    ),
]


def get_view_content(viewdef: ViewDefinition) -> str:
    if not VIEW_PATH.is_dir():
        raise Exception("View directory: {} does not exist".format(VIEW_PATH))

    view_full_path = VIEW_PATH / Path(viewdef.filepath)

    if not view_full_path.is_file():
        raise Exception("View {} not found in view path:".format(view_full_path))

    view_content: str = ""

    with view_full_path.open() as fh:
        view_content = fh.read()

    return view_content


POSTGIS_VIEWS = ["geography_columns", "geometry_columns", "raster_columns", "raster_overviews"]


def purge_views() -> None:
    """Remove views that aren't in the view table"""

    engine = get_database_engine()

    all_views_query = get_all_views_query()
    all_views = []

    with engine.connect() as c:
        result = list(c.execute(all_views_query))

    # Dont drop postgis or mapped views
    all_views = [i[0] for i in result if i[0] not in POSTGIS_VIEWS + [i.name for i in _VIEW_MAP]]

    for view_name in all_views:

        with engine.connect() as c:
            c.execution_options(isolation_level="AUTOCOMMIT")

            query = "drop materialized view if exists {} cascade;".format(view_name)

            logger.info("Dropping view {}".format(view_name))
            logger.debug(query)

            try:
                c.execute(query)
            except Exception as e:
                logger.error("Error dropping view: {}".format(e))


def init_database_views() -> None:
    """ Initialize all the database view """

    engine = get_database_engine()

    views_sorted_by_priority = list(sorted(_VIEW_MAP, key=attrgetter("priority")))

    for view in views_sorted_by_priority:
        logger.info("Initializing view {}".format(view.name))

        with engine.connect() as c:
            c.execution_options(isolation_level="AUTOCOMMIT")

            # drop
            drop_query = get_query_drop_view(view)

            logger.debug(drop_query)

            try:
                c.execute(drop_query)
            except Exception as e:
                logger.warn("Could not drop view {}".format(view.name))

            # create
            create_query = get_view_content(view)
            logger.debug(create_query)

            c.execute(create_query)

            # index
            index_create_query = get_view_unique_index_query(view)

            if index_create_query:
                logger.debug(index_create_query)

                try:
                    c.execute(index_create_query)
                except Exception as e:
                    logger.error("Error creating index: {}".format(e))

    return None


def init_aggregation_policies() -> None:
    """ Initializes the continuous aggregation policies """

    # @TODO check what exists with query

    engine = get_database_engine()

    for view in _VIEW_MAP:
        if not view.aggregation_policy:
            logging.debug("Skipping {}".format(view.name))
            continue

        with engine.connect() as c:

            drop_query = remove_continuous_aggregation_query(view)

            try:
                logger.debug(drop_query)
                c.execute(drop_query)
            except Exception as e:
                logger.warn("Could not drop continuous aggregation query: {}".format(view.name))
                pass

            create_query = create_continuous_aggregation_query(view)

            logger.debug(create_query)

            try:
                c.execute(create_query)
            except Exception as e:
                logger.warn("Could not create continuous aggregation query: {}".format(e))



# pylint: disable=no-self-argument
"""
OpenNEM Energy Tools

"""
from decimal import Decimal, getcontext
from operator import attrgetter
from sys import maxsize
from typing import Callable, Generator, List, Union

import numpy as np

from opennem.schema.core import BaseConfig

MAX_INTERVALS = maxsize - 1

context = getcontext()
context.prec = 9


class Point(BaseConfig):
    x: Decimal
    y: Decimal


def trapozedoid(p1: Point, p2: Point) -> float:
    """Trapezoidal area between two points"""
    width = p2.y - p1.y
    area = p1.x * width + ((p2.x - p1.x) * width) / 2

    return float(area)


def zero_nulls(number: Union[int, float, None]) -> float:
    if not number:
        return 0.0

    return number


def list_chunk(series: List, chunk_size: int) -> Generator[List[Union[float, int]], None, None]:
    for i in range(0, len(series), chunk_size):
        start_chunk = i - 1 if i > 0 else 0
        end_chunk = i + chunk_size + 1

        yield np.array(series[start_chunk:end_chunk])


def energy_sum(
    series: List[Union[int, float, None]],
    bucket_size_minutes: int,
    auc_function: Callable = trapozedoid,
) -> float:
    """Calcualte the energy sum of a series for an interval
    using the area under the curve
    """

    number_intervals = len(series) - 1
    interval_size = bucket_size_minutes / len(series)

    if len(series) < 1:
        raise Exception("Requires at least one value in the series")

    if bucket_size_minutes <= 0 or bucket_size_minutes > maxsize:
        raise Exception("Invalid bucket size")

    if bucket_size_minutes % 1 != 0:
        raise Exception("Not a round interval size in minutes")

    if number_intervals < 1 or number_intervals >= MAX_INTERVALS:
        raise Exception("Invalid number of intervals")

    # 0 out all nulls
    series = [zero_nulls(n) for n in series]

    y_series = [i * interval_size for i in range(len(series))]

    # build point objects from y_series and zeroed x values
    series_points = [Point(x=series[i], y=y) for i, y in enumerate(y_series)]

    # sort by y value ascending 0 ... N
    series_points.sort(key=attrgetter("y"))

    area = 0.0
    p1 = None
    p2 = None

    for i in range(number_intervals):
        p1 = series_points[i]
        p2 = series_points[i + 1]

        area += auc_function(p1, p2)

    # convert back to hours
    area = area / 60

    return area



import json
from typing import List

from opennem.controllers.stations import get_stations
from opennem.exporter.csv import stations_csv_records
from opennem.exporter.encoders import OpenNEMJSONEncoder
from opennem.schema.opennem import StationSchema


def stations_json_serialize(stations: List[StationSchema]):

    json_records = stations_csv_records(stations)

    json_serialized = json.dumps(
        json_records, cls=OpenNEMJSONEncoder, indent=4
    )

    return json_serialized



import re
from datetime import datetime, timedelta
from typing import Union

from datedelta import datedelta

TIME_INTERVALS = {
    "s": "seconds",
    "m": "minutes",
    "h": "hours",
    "d": "days",
    "w": "weeks",
}

DATE_INTERVALS = {
    "M": "months",
    "Y": "years",
}

VALID_INTERVALS = list(TIME_INTERVALS.keys()) + list(DATE_INTERVALS.keys())

VALID_INTERVAL_OPTIONS = ", ".join(VALID_INTERVALS)

_interval_parser = re.compile(r"(?P<value>\d+)(?P<unit>[smhdwMY]?)", re.IGNORECASE)


def get_human_interval(interval_human: str) -> Union[timedelta, datedelta]:
    """Parses a human interval like 1m or 1Y into a timedelta or datedelta"""
    interval_match = re.match(_interval_parser, interval_human.strip())

    if not interval_match:
        raise Exception("Not a valid interval: {}".format(interval_human))

    unit = interval_match.group("unit")
    value = int(interval_match.group("value"))

    assert unit in VALID_INTERVALS, f"Not a valid unit: {unit}. One of: {VALID_INTERVAL_OPTIONS}"
    assert value > 0, f"Not a valid number of intervals: {value}"

    if unit in TIME_INTERVALS.keys():
        return timedelta(**{TIME_INTERVALS[unit]: value})

    if unit in DATE_INTERVALS.keys():
        return datedelta(**{DATE_INTERVALS[unit]: value})

    raise Exception("Not a valid interval: {}".format(interval_human))


def add_human_inerval(dt: datetime, interval_human: str) -> datetime:
    """Add a human interval to a datetime"""
    return dt + get_human_interval(interval_human)



# pylint: disable=no-member
"""
mv_facility_all material view migration

@NOTE this was nullified to speed up deploys
as the next migration recreates with updates

Revision ID: 0a7eaccf205e
Revises: 43f1d539490f
Create Date: 2021-01-12 18:53:42.852651

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "0a7eaccf205e"
down_revision = "43f1d539490f"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        """
        drop materialized view if exists mv_facility_all;
    """
    )


def downgrade() -> None:
    op.execute("drop materialized view if exists mv_facility_all;")



from opennem.pipelines.bulk_insert import BulkInsertPipeline
from opennem.pipelines.csv import RecordsToCSVPipeline
from opennem.pipelines.nem.opennem import NemwebUnitScadaOpenNEMStorePipeline
from opennem.spiders.nemweb import NemwebSpider


class NemwebLatestDispatchScada(NemwebSpider):
    name = "au.nem.latest.dispatch_scada"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/Dispatch_SCADA/"
    limit = 2

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebCurrentDispatchScada(NemwebSpider):
    name = "au.nem.current.dispatch_scada"
    start_url = "http://www.nemweb.com.au/Reports/CURRENT/Dispatch_SCADA/"
    limit = 0
    skip = 2

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )


class NemwebArchiveDispatchScada(NemwebSpider):
    name = "au.nem.archive.dispatch_scada"
    start_url = "http://www.nemweb.com.au/Reports/ARCHIVE/Dispatch_SCADA/"

    pipelines_extra = set(
        [
            NemwebUnitScadaOpenNEMStorePipeline,
            BulkInsertPipeline,
            RecordsToCSVPipeline,
        ]
    )

    # Archives tend to contain large zips of embedded zips so throttle
    # to limit memory use
    custom_settings = {
        "CONCURRENT_REQUESTS": 4,
        "CONCURRENT_ITEMS": 1,
    }



# pylint: disable=no-member
"""Emission factor co2 to facility

Revision ID: 0bd670d87da9
Revises: 69449fda6381
Create Date: 2020-11-13 02:48:57.159985

"""
import sqlalchemy as sa
from alembic import op

revision = "0bd670d87da9"
down_revision = "69449fda6381"
branch_labels = None
depends_on = None


def upgrade():
    op.add_column(
        "facility",
        sa.Column("emissions_factor_co2", sa.Numeric(), nullable=True),
    )


def downgrade():
    op.drop_column("facility", "emissions_factor_co2")



# pylint: disable=no-member
"""
Remove mv_network_energy_hour

Revision ID: 9c7482a0fc9a
Revises: 977f219c8563
Create Date: 2021-01-19 16:19:01.151500

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "9c7482a0fc9a"
down_revision = "977f219c8563"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("drop view if exists mv_network_energy_hour cascade;")


def downgrade() -> None:
    pass



import json
import logging
from datetime import datetime

from scrapy.exceptions import DropItem
from sqlalchemy.exc import IntegrityError
from sqlalchemy.sql import text

from opennem.db.models.opennem import Facility, Participant
from opennem.pipelines import DatabaseStoreBase
from opennem.utils.pipelines import check_spider_pipeline

logger = logging.getLogger(__name__)



# pylint: disable=no-member
"""
create mv_wem_facility_power_30min

Revision ID: e6bb59b27a23
Revises: 26ecffa41c19
Create Date: 2021-01-22 02:08:14.124716

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "e6bb59b27a23"
down_revision = "26ecffa41c19"
branch_labels = None
depends_on = None


stmt_drop = "drop materialized view if exists mv_wem_facility_power_30min cascade;"

stmt = """
    CREATE MATERIALIZED VIEW mv_wem_facility_power_30min WITH (timescaledb.continuous) AS
    select
        time_bucket('30 minutes', fs.trading_interval) as trading_interval,
        fs.facility_code,
        max(fs.generated) as facility_power
    from facility_scada fs
    where fs.network_id = 'WEM'
    group by
        1,
        fs.facility_code;
"""


def upgrade() -> None:
    with op.get_context().autocommit_block():
        op.execute(stmt_drop)
        op.execute(stmt)


def downgrade() -> None:
    op.execute(stmt_drop)



from opennem.api.geo.router import geo_facilities_api
from opennem.db import SessionLocal
from opennem.exporter.aws import write_to_s3


def export_facility_geojson() -> None:
    session = SessionLocal()

    facility_geo = geo_facilities_api(only_approved=True, session=session)

    write_to_s3("/v3/geo/au_facilities.json", facility_geo.json())


if __name__ == "__main__":
    export_facility_geojson()



# pylint: disable=no-member
"""BOM station geom

Revision ID: 8bf30c37cb85
Revises: d45e70a09231
Create Date: 2020-10-06 14:27:28.910935

"""
import geoalchemy2
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "8bf30c37cb85"
down_revision = "d45e70a09231"
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "bom_station",
        sa.Column(
            "geom",
            geoalchemy2.types.Geometry(
                geometry_type="POINT",
                srid=4326,
                spatial_index=False,
                from_text="ST_GeomFromEWKT",
                name="geometry",
            ),
            nullable=True,
        ),
    )
    op.create_index(
        "idx_bom_station_geom",
        "bom_station",
        ["geom"],
        unique=False,
        postgresql_using="gist",
    )
    op.execute(
        "update bom_station set geom=ST_SetSRID(ST_MakePoint(lng, lat), 4326);"
    )
    op.drop_column("bom_station", "lng")
    op.drop_column("bom_station", "lat")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column(
        "bom_station",
        sa.Column("lat", sa.NUMERIC(), autoincrement=False, nullable=True),
    )
    op.add_column(
        "bom_station",
        sa.Column("lng", sa.NUMERIC(), autoincrement=False, nullable=True),
    )
    op.drop_index("idx_bom_station_geom", table_name="bom_station")
    op.drop_column("bom_station", "geom")
    # ### end Alembic commands ###



"""
    Queries for network data

    @TODO make these pluggable
    @TODO use sqlalchemy text() compiled queries
"""

from datetime import datetime, timedelta
from textwrap import dedent
from typing import List, Optional

from fastapi.exceptions import HTTPException
from starlette import status

from opennem.api.stats.controllers import get_scada_range
from opennem.api.stats.schema import ScadaDateRange
from opennem.core.networks import network_from_network_region
from opennem.core.normalizers import normalize_duid
from opennem.schema.network import NetworkSchema
from opennem.schema.time import TimeInterval, TimePeriod


def duid_in_case(facility_codes: List[str]) -> str:
    return ",".join(["'{}'".format(i) for i in map(normalize_duid, facility_codes)])


def power_facility_query(
    facility_codes: List[str],
    network: NetworkSchema,
    period: TimePeriod,
    interval: Optional[TimeInterval] = None,
    date_range: Optional[ScadaDateRange] = None,
) -> str:
    timezone = network.get_timezone(postgres_format=True)

    if not date_range:
        date_range = get_scada_range(network=network, facilities=facility_codes)

    timezone = network.timezone_database

    __query = """
        select
            t.trading_interval at time zone '{timezone}',
            coalesce(avg(t.facility_power), 0),
            t.facility_code
        from (
            select
                time_bucket_gapfill('{trunc}', fs.trading_interval) AS trading_interval,
                coalesce(
                    avg(fs.generated), 0
                ) as facility_power,
                fs.facility_code
            from facility_scada fs
            join facility f on fs.facility_code = f.code
            where
                fs.trading_interval <= '{date_max}' and
                fs.trading_interval > '{date_min}' and
                fs.facility_code in ({facility_codes_parsed})
            group by 1, 3
        ) as t
        group by 1, 3
        order by 1 desc
    """

    if not interval:
        interval = network.get_interval()

    if not date_range:
        raise Exception("Require a date range for query")

    if not interval:
        raise Exception("Require an interval")

    date_max = date_range.get_end()
    date_min = date_range.get_start()

    if period:
        date_min = date_range.get_end() - timedelta(minutes=period.period)

    query = __query.format(
        facility_codes_parsed=duid_in_case(facility_codes),
        trunc=interval.interval_sql,
        period=period.period_sql,
        timezone=timezone,
        date_max=date_max,
        date_min=date_min,
    )

    return query


def power_network(
    network_code: str,
    interval: TimeInterval,
    period: TimePeriod,
) -> str:

    scale = 1

    network = network_from_network_region(network_code)
    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """
        with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.facility_code as facility_code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.facility_code,
                coalesce(avg(generated), 0) as generated
                from facility_scada fs
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 desc
    """

    query = __query.format(
        network_code=network_code,
        trunc=interval.trunc,
        interval=interval.interval_human,
        interval_remainder=interval.get_sql_join(timezone=network.timezone_database),
        period=period.period_sql,
        scale=scale,
        timezone=timezone,
    )

    return query


def power_network_fueltech(
    network: NetworkSchema,
    interval: TimeInterval,
    period: TimePeriod,
    network_region: Optional[str] = None,
    scada_range: Optional[ScadaDateRange] = None,
) -> str:

    timezone = network.get_timezone(postgres_format=True)

    if not timezone:
        timezone = "UTC"

    __query = """
        SET SESSION TIME ZONE '{timezone}';

        select
            t.trading_interval,
            sum(t.facility_power),
            t.fueltech_code
        from (
            select
                time_bucket_gapfill('{trunc}', trading_interval) AS trading_interval,
                coalesce(
                    avg(fs.generated), 0
                ) as facility_power,
                fs.facility_code,
                ft.code as fueltech_code
            from facility_scada fs
            join facility f on fs.facility_code = f.code
            join fueltech ft on f.fueltech_id = ft.code
            where
                fs.trading_interval <= {date_end}
                and fs.trading_interval >= {date_end}::timestamp - '{period}'::interval
                and fs.network_id = '{network_code}'
                and f.fueltech_id is not null
                {network_region_query}
            group by 1, 3, 4
        ) as t
        group by 1, 3
        order by 1 desc
    """

    network_region_query = ""

    if network_region:
        network_region_query = f"and f.network_region='{network_region}'"

    date_end = "now()"

    if scada_range:
        date_end = scada_range.get_end_sql()

    query = __query.format(
        network_code=network.code,
        trunc=interval.interval_sql,
        period=period.period_sql,
        network_region_query=network_region_query,
        timezone=timezone,
        date_end=date_end,
    )

    return query


def energy_facility(
    facility_codes: List[str],
    network_code: str,
    interval: TimeInterval,
    period: TimePeriod,
) -> str:

    network = network_from_network_region(network_code)
    timezone = network.timezone_database

    __query = """
        with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.facility_code as facility_code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.facility_code,
                coalesce(sum(fs.eoi_quantity), 0.0) / {scale} as generated
                from facility_scada fs
                where
                    fs.facility_code in ({facility_codes_parsed})
                    and fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 asc"""

    query = __query.format(
        facility_codes_parsed=duid_in_case(facility_codes),
        network_code=network_code,
        trunc=interval.trunc,
        interval=interval.interval_sql,
        interval_remainder=interval.get_sql_join(timezone=network.timezone_database),
        period=period.period_sql,
        scale=network.intervals_per_hour,
        timezone=timezone,
    )

    return query


def energy_facility_query(
    facility_codes: List[str],
    network: NetworkSchema,
    period: TimePeriod,
    interval: Optional[TimeInterval] = None,
) -> str:
    """
    Get Energy for a list of facility codes
    """

    __query = """
    select
        date_trunc('{trunc}', t.trading_interval at time zone '{timezone}') as trading_day,
        t.code,
        sum(t.energy) as fueltech_energy,
        sum(t.market_value) as fueltech_market_value,
        sum(t.emissions) as fueltech_emissions
    from mv_facility_all t
    where
        t.trading_interval <= '{date_max}' and
        t.trading_interval >= '{date_min}' and
        t.code in ({facility_codes_parsed})
    group by 1, 2
    order by
        trading_day desc;
    """

    timezone = network.timezone_database
    offset = network.get_timezone(postgres_format=True)

    date_range: ScadaDateRange = get_scada_range(network=network, facilities=facility_codes)

    if not interval:
        interval = network.get_interval()

    if not date_range:
        raise Exception("Require a date range for query")

    if not period:
        raise Exception("Require a period")

    if not interval:
        interval = network.get_interval()

    if not interval:
        raise Exception("Require an interval")

    trunc = interval.trunc

    date_max = date_range.get_end().replace(hour=23, minute=59, second=59) - timedelta(days=1)
    date_min = date_range.get_start()

    if period.period_human == "1M":
        date_min = date_range.get_end() - timedelta(minutes=period.period)
    elif period.period_human == "1Y":
        # might have to do +offset times
        year = datetime.now().year - 1
        date_min = date_max.replace(year=year)
    elif period.period_human in ["7d", "5Y", "10Y"]:
        date_min = date_range.get_end() - timedelta(minutes=period.period)

    # elif period.period_human == "all":
    # else:
    # date_min = date_range.get_end() - timedelta(minutes=period.period)

    query = dedent(
        __query.format(
            facility_codes_parsed=duid_in_case(facility_codes),
            trunc=trunc,
            date_max=date_max,
            date_min=date_min,
            timezone=timezone,
        )
    )
    return query


def energy_network(
    network: NetworkSchema,
    interval: TimeInterval,
    period: TimePeriod,
) -> str:

    timezone = network.timezone_database

    __query = """
        with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.facility_code as facility_code
        from intervals i
        left outer join
            (select
                date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                fs.facility_code,
                coalesce(sum(fs.eoi_quantity), 0.0) / {scale} as generated
                from facility_scada fs
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 asc"""

    query = __query.format(
        network_code=network.code,
        trunc=interval.trunc,
        interval=interval.interval_sql,
        interval_remainder=interval.get_sql_join(timezone=network.timezone_database),
        period=period.period_sql,
        scale=network.intervals_per_hour,
        timezone=timezone,
    )

    return query


def energy_network_fueltech(
    network: NetworkSchema,
    interval: TimeInterval,
    period: TimePeriod,
    network_region: str = None,
) -> str:

    timezone = network.timezone_database

    if not timezone:
        timezone = "UTC"

    __query = """
        with intervals as (
            select generate_series(
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}') - '{period}'::interval,
                date_trunc('{trunc}', now() AT TIME ZONE '{timezone}'),
                '{interval}'::interval
            )::timestamp as interval
        )

        select
            i.interval as trading_day,
            fs.generated,
            fs.code
        from intervals i
        left outer join
            (
                select
                    date_trunc('{trunc}', fs.trading_interval AT TIME ZONE '{timezone}')::timestamp  {interval_remainder} as interval,
                    ft.code,
                    coalesce(sum(fs.eoi_quantity), 0) / {scale} as generated
                from facility_scada fs
                join facility f on fs.facility_code = f.code
                join fueltech ft on f.fueltech_id = ft.code
                where
                    fs.trading_interval > now() AT TIME ZONE '{timezone}' - '{period}'::interval
                    and fs.network_id = '{network_code}'
                    and f.fueltech_id is not null
                    {network_region_query}
                group by 1, 2
            ) as fs on fs.interval = i.interval
        order by 1 desc, 2 desc"""

    network_region_query = ""

    if network_region:
        network_region_query = f"and f.network_region='{network_region}'"

    query = __query.format(
        network_code=network.code,
        trunc=interval.trunc,
        interval=interval.interval_sql,
        interval_remainder=interval.get_sql_join(timezone=network.timezone_database),
        period=period.period_sql,
        scale=network.intervals_per_hour,
        network_region_query=network_region_query,
        timezone=timezone,
    )

    return query


def energy_network_fueltech_year(
    network: NetworkSchema,
    interval: TimeInterval,
    year: int,
    network_region: str = None,
    scada_range: ScadaDateRange = None,
) -> str:
    """
    Get Energy for a network or network + region
    based on a year
    """

    timezone = network.get_timezone(postgres_format=True)

    if not timezone:
        timezone = "UTC"

    year_max = "'{}-12-31'".format(year)

    if year == datetime.now().year:
        year_max = scada_range.get_end_sql(as_date=False)

    __query = """
        SET SESSION TIME ZONE '{timezone}';

        select
            t.trading_interval,
            sum(t.facility_energy),
            t.fueltech_code
        from (
            select
                time_bucket_gapfill('{trunc}', trading_interval) AS trading_interval,
                energy_sum(fs.generated, '{trunc}') * interval_size('1 day', count(fs.generated)) / 1000 as facility_energy,
                f.code,
                ft.code as fueltech_code
            from facility_scada fs
            join facility f on fs.facility_code = f.code
            join fueltech ft on f.fueltech_id = ft.code
            where
                fs.trading_interval >= '{year}-01-01'
                and fs.trading_interval <= {year_max}
                and fs.network_id = '{network_code}'
                and f.fueltech_id is not null
                {network_region_query}
            group by 1, 3, 4
        ) as t
        group by 1, 3
        order by 1 desc;
    """

    network_region_query = ""

    if network_region:
        network_region_query = f"and f.network_region='{network_region}'"

    query = __query.format(
        network_code=network.code,
        trunc=interval.interval_sql,
        year=year,
        year_max=year_max,
        scale=network.intervals_per_hour,
        network_region_query=network_region_query,
        timezone=timezone,
    )

    return query


def energy_network_fueltech_all(
    network: Optional[NetworkSchema],
    network_region: Optional[str],
    scada_range: ScadaDateRange,
):
    timezone = "AEST"

    if network:
        timezone = network.get_timezone(postgres_format=True)

    __query = """

        SET SESSION TIME ZONE '{timezone}';

        select
            date_trunc('month', t.trading_interval),
            sum(t.facility_energy),
            t.fueltech_code
        from (
            select
                time_bucket_gapfill('1 day', trading_interval) AS trading_interval,
                energy_sum(fs.generated, '1 day') * interval_size('1 day', count(fs.generated)) / 1000 as facility_energy,
                f.code,
                ft.code as fueltech_code
            from facility_scada fs
            join facility f on fs.facility_code = f.code
            join fueltech ft on f.fueltech_id = ft.code
            where
                fs.trading_interval >= {scada_min}
                and fs.trading_interval <= {scada_max}
                and f.fueltech_id is not null
                {network_query}
                {network_region_query}
            group by 1, 3, 4
        ) as t
        group by 1, 3
        order by 1 desc;
    """

    network_query = ""
    network_region_query = ""

    if network:
        network_query = f"and fs.network_id = '{network.code}' "

    if network_region:
        network_region_query = f"and f.network_region='{network_region}' "

    query = __query.format(
        network_code=network.code,
        network_query=network_query,
        network_region_query=network_region_query,
        scada_min=scada_range.get_start_sql(as_date=True),
        scada_max=scada_range.get_end_sql(as_date=True),
        timezone=timezone,
    )

    return query


def price_network_region(
    network: NetworkSchema,
    network_region_code: str,
    interval: TimeInterval,
    period: TimePeriod,
    scada_range: ScadaDateRange,
    year: Optional[int] = None,
) -> str:

    timezone = network.get_timezone(postgres_format=True)

    if not timezone:
        timezone = "UTC"

    __query = """
        SET SESSION TIME ZONE '{timezone}';

        select
            time_bucket_gapfill('{trunc}', bs.trading_interval) AS trading_interval,
            bs.network_region,
            coalesce(avg(bs.price), 0) as price
        from balancing_summary bs
        where
            bs.trading_interval >= {date_min_query}
            and bs.trading_interval <= {scada_max}
            {network_query}
            {network_region_query}
        group by 1, 2
        order by 1 desc
    """

    network_query = ""
    network_region_query = ""

    if network:
        network_query = f"and bs.network_id = '{network.code}' "

    if network_region_code:
        network_region_query = f"and bs.network_region='{network_region_code}' "

    date_min_query = ""

    if period:
        date_min_query = "{scada_max}::timestamp - interval '{period}'::interval".format(
            scada_max=scada_range.get_end_sql(), period=period.period_sql
        )

    if year:
        date_min_query = "'{year}-01-01'::date ".format(year=year)

    if not period and not year:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail="Require one of period or year",
        )

    query = __query.format(
        trunc=interval.interval_sql,
        timezone=timezone,
        network_query=network_query,
        network_region_query=network_region_query,
        scada_max=scada_range.get_end_sql(),
        date_min_query=date_min_query,
    )

    return query


def price_network_monthly(
    network: Optional[NetworkSchema],
    network_region_code: Optional[str],
    scada_range: ScadaDateRange,
):
    timezone = "AEST"

    networks = [network.code]

    if network.code == "au":
        networks = ["WEM", "NEM"]

    if network:
        timezone = network.get_timezone(postgres_format=True)

    __query = """

        SET SESSION TIME ZONE '{timezone}';

        select
            date_trunc('month', t.trading_interval),
            avg(t.price),
            t.network_id
        from (
            select
                time_bucket_gapfill('1 day', trading_interval) AS trading_interval,
                avg(bs.price) as price,
                bs.network_id as network_id
            from balancing_summary bs
            where
                bs.trading_interval >= {scada_min}
                and bs.trading_interval <= {scada_max}
                and bs.network_id IN ({network_codes})
            group by 1, 3
        ) as t
        group by 1, 3
        order by 1 desc;
    """

    network_codes = ""

    for network_code in networks:
        network_codes = ", ".join("'{}'".format(network_code.upper()))

    query = __query.format(
        network_codes=network_codes,
        scada_min=scada_range.get_start_sql(as_date=True),
        scada_max=scada_range.get_end_sql(as_date=True),
        timezone=timezone,
    )

    return query



# pylint: disable=no-member
"""
Update fueltechs

Revision ID: 6e45ee9557c2
Revises: b2c0f8c9d5db
Create Date: 2021-02-02 16:16:37.583127

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "6e45ee9557c2"
down_revision = "b2c0f8c9d5db"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute(
        "update facility set fueltech_id='wind' where code IN ('LBBL1', 'BLAIRFOX_BEROSRD_WF1')"
    )


def downgrade() -> None:
    pass



"""
OpenNEM Stat Set Loader

Takes a stat set or data history from exports in Dict form and returns
schemas.

"""
from typing import Any, Dict

from opennem.api.stats.schema import OpennemDataSet


def load_statset(stat_set: Dict[str, Any]) -> OpennemDataSet:
    """ """

    ds = OpennemDataSet.parse_obj(stat_set)

    return ds



# pylint: disable=no-member
"""
Update continuous integration policies for mv_facility_energy_hour

Revision ID: b2c0f8c9d5db
Revises: 60ff840c16c5
Create Date: 2021-02-02 02:09:35.724514

"""
from alembic import op
from sqlalchemy.exc import ProgrammingError

# revision identifiers, used by Alembic.
revision = "b2c0f8c9d5db"
down_revision = "60ff840c16c5"
branch_labels = None
depends_on = None

VIEWS = [
    {
        "name": "mv_facility_energy_hour",
        "interval": "30 minutes",
        "start_offset": "6 hours",
        "end_offset": "1 hour",
    },
]

query_drop = "SELECT remove_continuous_aggregate_policy('{view_name}')"

query_create = """SELECT add_continuous_aggregate_policy('{view_name}',
    start_offset => INTERVAL '{start_offset}',
    end_offset => '{end_offset}',
    schedule_interval => INTERVAL '{schedule_size}');
"""


def upgrade() -> None:
    for v in VIEWS:
        q_drop = query_drop.format(view_name=v["name"])
        q_create = query_create.format(
            view_name=v["name"],
            schedule_size=v["interval"],
            start_offset=v["start_offset"],
            end_offset=v["end_offset"],
        )

        with op.get_context().autocommit_block():
            try:
                op.execute(q_drop)
            except ProgrammingError:
                pass

        with op.get_context().autocommit_block():
            op.execute(q_create)


def downgrade() -> None:
    pass



# pylint: disable=no-member
"""
update Barcaldine solar farm capacity

Revision ID: 57c6030dcb87
Revises: fcb509301bda
Create Date: 2021-02-09 23:29:05.629873

"""
from alembic import op

# revision identifiers, used by Alembic.
revision = "57c6030dcb87"
down_revision = "fcb509301bda"
branch_labels = None
depends_on = None


def upgrade() -> None:
    op.execute("update facility set capacity_registered=20 where code='BARCSF1';")


def downgrade() -> None:
    pass



# import json
from typing import Any, Dict, Generator

import scrapy

from opennem.pipelines.nem.summary import NemwebSummaryPipeline


class NemSummarySpider(scrapy.Spider):
    name = "au.nem.summary"
    start_urls = ["https://www.aemo.com.au/aemo/apps/api/report/ELEC_NEM_SUMMARY"]

    pipelines_extra = set(
        [
            NemwebSummaryPipeline,
        ]
    )

    def parse(self, response: Any) -> Generator[Dict, None, None]:
        yield {"content": response.json()}



"""
OpenNEM v2 formatted output exports as Pydantic schemas

"""

from datetime import date, datetime
from typing import List, Optional, Tuple, Union

from pydantic import validator

# from opennem.api.stats.schema import data_validate
from opennem.schema.core import BaseConfig
from opennem.utils.interval import get_human_interval

# from pydantic import validator


def strip_timezone(dt: datetime) -> datetime:
    return dt.replace(tzinfo=None)


def fix_v2_date(date_str: str) -> str:
    return date_str.replace("+1000", "+10:00")


class OpennemDataHistoryV2(BaseConfig):
    start: Union[datetime, date]
    last: Union[datetime, date]
    interval: str
    data: List[Optional[float]]

    # validators
    _start_date = validator("start", allow_reuse=True, pre=True)(fix_v2_date)
    _end_date = validator("last", allow_reuse=True, pre=True)(fix_v2_date)

    def values(self) -> List[Tuple[datetime, float]]:
        interval_obj = get_human_interval(self.interval)
        dt = self.start

        # return as list rather than generate
        xy_values = []

        for v in self.data:
            dt = dt + interval_obj
            xy_values.append((strip_timezone(dt), v))

        return xy_values


class OpennemDataV2(BaseConfig):
    id: str
    region: str
    type: Optional[str]
    units: str
    fuel_tech: Optional[str]

    history: OpennemDataHistoryV2
    forecast: Optional[OpennemDataHistoryV2]


class OpennemDataSetV2(BaseConfig):
    version: str = "v2"
    data: List[OpennemDataV2]

    def get_id(self, id: str) -> Optional[OpennemDataV2]:
        _ds = list(filter(lambda x: x.id == id, self.data))

        if len(_ds) < 1:
            return None

        return _ds.pop()



# pylint: disable=no-name-in-module
# pylint: disable=no-self-argument
# pylint: disable=no-member
import platform

from huey import PriorityRedisHuey, crontab

from opennem.db.tasks import refresh_material_views
from opennem.settings import settings
from opennem.workers.energy import run_energy_update

# Py 3.8 on MacOS changed the default multiprocessing model
if platform.system() == "Darwin":
    import multiprocessing

    try:
        multiprocessing.set_start_method("fork")
    except RuntimeError:
        # sometimes it has already been set by
        # other libs
        pass

redis_host = None

if settings.cache_url:
    redis_host = settings.cache_url.host

huey = PriorityRedisHuey("opennem.scheduler.db", host=redis_host)


@huey.periodic_task(crontab(hour="*/12"))
@huey.lock_task("db_refresh_material_views")
def db_refresh_material_views() -> None:
    if settings.workers_db_run:
        refresh_material_views()


@huey.periodic_task(crontab(hour="*/12"))
@huey.lock_task("db_refrehs_energies_yesterday")
def db_refrehs_energies_yesterday() -> None:
    if settings.workers_db_run:
        run_energy_update()



from datetime import datetime

from opennem.api.schema import ApiBase
from opennem.core.networks import NetworkSchema


class ScraperStats(ApiBase):
    scada_intervals: int
    scada_min: datetime
    scada_max: datetime


class ScraperStatsResult(ApiBase):
    network: NetworkSchema
    stats: ScraperStats



"""
    Generates the map of queries and paths to export to S3 buckets
    for opennem.org.au

"""
from __future__ import annotations

import logging
from datetime import datetime
from enum import Enum
from typing import List, Optional

from pydantic import BaseModel

from opennem.api.stats.controllers import ScadaDateRange, get_scada_range
from opennem.api.time import human_to_interval, human_to_period
from opennem.core.network_region_bom_station_map import get_network_region_weather_station
from opennem.core.networks import (
    NetworkAPVI,
    NetworkAU,
    NetworkNEM,
    NetworkSchema,
    NetworkWEM,
    network_from_network_code,
)
from opennem.db import SessionLocal
from opennem.db.models.opennem import Network
from opennem.schema.time import TimeInterval, TimePeriod
from opennem.utils.dates import date_range_from_week, week_series
from opennem.utils.version import VersionPart, get_version

logger = logging.getLogger(__name__)

VERSION_MAJOR = get_version(version_part=VersionPart.MAJOR)
STATS_FOLDER = "stats"


class StatType(Enum):
    power = "power"
    energy = "energy"
    gov = "gov"


class PriorityType(Enum):
    live = 1
    daily = 2
    monthly = 3
    history = 4


def priority_from_name(priority_name: str) -> PriorityType:
    # @TODO this doesn't feel right
    if priority_name == "live":
        return PriorityType.live
    if priority_name == "daily":
        return PriorityType.daily
    if priority_name == "monthly":
        return PriorityType.monthly
    if priority_name == "history":
        return PriorityType.history

    raise Exception("Could not find priority: {}".format(priority_name))


class StatExport(BaseModel):
    """Defines an export map - a set of variables that produces a JSON export"""

    stat_type: StatType = StatType.energy
    priority: PriorityType = PriorityType.live
    country: str
    network: NetworkSchema
    networks: Optional[List[NetworkSchema]]
    network_region: Optional[str]
    network_region_query: Optional[str]
    date_range: Optional[ScadaDateRange]
    bom_station: Optional[str]
    year: Optional[int]
    week: Optional[int]
    period: Optional[TimePeriod]
    interval: TimeInterval
    file_path: Optional[str]

    @property
    def path(self) -> str:
        _path_components = [
            f"v{VERSION_MAJOR}",
            STATS_FOLDER,
            self.country,
            self.network.code,
        ]

        if self.network_region:
            _path_components.append(self.network_region)

        _path_components.append(self.stat_type.value)

        # only show the period when it's not explicitly a year
        if self.period and not self.year:
            _path_components.append(self.period.period_human)

        if self.week:
            _path_components.append("week")

        if self.year:
            _path_components.append(str(self.year))

        if self.week:
            _path_components.append(str(self.week))

        dir_path = "/".join([str(i) for i in _path_components])

        return "{}.json".format(dir_path)


class StatMetadata(BaseModel):
    """Defines a set of export maps with methods to filter"""

    date_created: datetime
    version: Optional[str]
    resources: List[StatExport]

    def get_by_stat_type(self, stat_type: StatType) -> StatMetadata:
        em = self.copy()
        em.resources = list(filter(lambda s: s.stat_type == stat_type, self.resources))
        return em

    def get_by_network_id(
        self,
        network_id: str,
    ) -> StatMetadata:
        em = self.copy()
        em.resources = list(filter(lambda s: s.network.code == network_id, self.resources))
        return em

    def get_by_network_region(
        self,
        network_region: str,
    ) -> StatMetadata:
        em = self.copy()
        em.resources = list(filter(lambda s: s.network_region == network_region, self.resources))
        return em

    def get_by_year(
        self,
        year: int,
    ) -> StatMetadata:
        em = self.copy()
        em.resources = list(filter(lambda s: s.year == year, self.resources))
        return em

    def get_by_years(
        self,
        years: List[int],
    ) -> StatMetadata:
        em = self.copy()
        em.resources = list(filter(lambda s: s.year in years, self.resources))
        return em

    def get_by_priority(self, priority: PriorityType) -> StatMetadata:
        em = self.copy()
        em.resources = list(
            filter(
                lambda s: s.priority == priority,
                self.resources,
            )
        )
        return em


def get_export_map() -> StatMetadata:
    """
    Generates a map of all export JSONs

    """
    session = SessionLocal()

    networks = session.query(Network).filter(Network.export_set.is_(True)).all()

    if not networks:
        raise Exception("No networks")

    countries = list(set([network.country for network in networks]))

    _exmap = []

    for country in countries:
        # @TODO derive this
        scada_range = get_scada_range(network=NetworkAU, networks=[NetworkNEM, NetworkWEM])

        if not scada_range:
            raise Exception("Require a scada range")

        export = StatExport(
            stat_type=StatType.power,
            priority=PriorityType.live,
            country=country,
            date_range=scada_range,
            network=NetworkAU,
            networks=[NetworkNEM, NetworkWEM],
            interval=NetworkAU.get_interval(),
            period=human_to_period("7d"),
        )

        _exmap.append(export)

        for year, week in week_series(scada_range.end, scada_range.start):
            export = StatExport(
                stat_type=StatType.power,
                priority=PriorityType.history,
                country=country,
                network=NetworkAU,
                networks=[NetworkNEM, NetworkWEM],
                year=year,
                week=week,
                interval=human_to_interval("1d"),
                period=human_to_period("7d"),
                date_range=date_range_from_week(year, week, NetworkAU),
            )
            _exmap.append(export)

        for year in range(
            datetime.now().year,
            scada_range.start.year - 1,
            -1,
        ):
            export = StatExport(
                stat_type=StatType.energy,
                priority=PriorityType.daily,
                country=country,
                date_range=scada_range,
                network=NetworkAU,
                networks=[NetworkNEM, NetworkWEM],
                year=year,
                interval=human_to_interval("1d"),
                period=human_to_period("1Y"),
            )
            _exmap.append(export)

        export = StatExport(
            stat_type=StatType.energy,
            priority=PriorityType.monthly,
            country=country,
            date_range=scada_range,
            network=NetworkAU,
            networks=[NetworkNEM, NetworkWEM],
            interval=human_to_interval("1M"),
            period=human_to_period("all"),
        )
        _exmap.append(export)

    for network in networks:
        network_schema = network_from_network_code(network.code)
        scada_range = get_scada_range(network=network_schema)
        bom_station = get_network_region_weather_station(network.code)

        export = StatExport(
            stat_type=StatType.power,
            priority=PriorityType.live,
            country=network.country,
            date_range=scada_range,
            network=network_schema,
            bom_station=bom_station,
            interval=network_schema.get_interval(),
            period=human_to_period("7d"),
        )

        if network.code == "WEM":
            export.networks = [NetworkWEM, NetworkAPVI]
            export.network_region_query = "WEM"

        _exmap.append(export)

        if not scada_range:
            raise Exception("Require a scada range")

        for year, week in week_series(scada_range.end, scada_range.start):
            export = StatExport(
                stat_type=StatType.power,
                priority=PriorityType.history,
                country=network.country,
                network=network_schema,
                year=year,
                week=week,
                interval=human_to_interval("1d"),
                period=human_to_period("7d"),
                date_range=date_range_from_week(year, week, NetworkAU),
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

        for year in range(
            datetime.now().year,
            scada_range.start.year - 1,
            -1,
        ):
            export = StatExport(
                stat_type=StatType.energy,
                priority=PriorityType.daily,
                country=network.country,
                date_range=scada_range,
                network=network_schema,
                bom_station=bom_station,
                year=year,
                period=human_to_period("1Y"),
                interval=human_to_interval("1d"),
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

        export = StatExport(
            stat_type=StatType.energy,
            priority=PriorityType.monthly,
            country=network.country,
            date_range=scada_range,
            network=network_schema,
            bom_station=bom_station,
            interval=human_to_interval("1M"),
            period=human_to_period("all"),
        )

        if network.code == "WEM":
            export.networks = [NetworkWEM, NetworkAPVI]
            export.network_region_query = "WEM"

        _exmap.append(export)

        # Skip cases like wem/wem where region is supurfelous
        if len(network.regions) < 2:
            continue

        for region in network.regions:
            scada_range = get_scada_range(network=network_schema, network_region=region)
            bom_station = get_network_region_weather_station(region.code)

            if not scada_range:
                raise Exception("Require a scada range")

            export = StatExport(
                stat_type=StatType.power,
                priority=PriorityType.live,
                country=network.country,
                date_range=scada_range,
                network=network_schema,
                network_region=region.code,
                bom_station=bom_station,
                period=human_to_period("7d"),
                interval=network_schema.get_interval(),
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

            for year in range(
                datetime.now().year,
                scada_range.start.year - 1,
                -1,
            ):
                export = StatExport(
                    stat_type=StatType.energy,
                    priority=PriorityType.daily,
                    country=network.country,
                    date_range=scada_range,
                    network=network_schema,
                    network_region=region.code,
                    bom_station=bom_station,
                    year=year,
                    period=human_to_period("1Y"),
                    interval=human_to_interval("1d"),
                )
                _exmap.append(export)

            export = StatExport(
                stat_type=StatType.energy,
                priority=PriorityType.monthly,
                country=network.country,
                date_range=scada_range,
                network=network_schema,
                network_region=region.code,
                bom_station=bom_station,
                period=human_to_period("all"),
                interval=human_to_interval("1M"),
            )

            if network.code == "WEM":
                export.networks = [NetworkWEM, NetworkAPVI]
                export.network_region_query = "WEM"

            _exmap.append(export)

    export_meta = StatMetadata(
        date_created=datetime.now(), version=get_version(), resources=_exmap
    )

    return export_meta


if __name__ == "__main__":
    pass



"""
Queries to drop views and create primary key indexes

"""

from textwrap import dedent
from typing import Optional

from opennem.db.views.utils import get_index_name

from .schema import ViewDefinition


def query_materialized(viewdef: ViewDefinition) -> str:
    return "materialized" if viewdef.materialized else ""


def get_query_drop_view(viewdef: ViewDefinition) -> Optional[str]:

    __query = "drop {materialized} view if exists {view_name} cascade;"

    query = __query.format(
        view_name=viewdef.name,
        materialized=query_materialized(viewdef),
    )

    return dedent(query)


def get_view_unique_index_query(viewdef: ViewDefinition) -> Optional[str]:

    if not viewdef.primary_key:
        return None

    __query = "CREATE UNIQUE INDEX IF NOT EXISTS {index_name} ON {view_name} ({index_keys});"

    index_keys = ", ".join(viewdef.primary_key)

    query = __query.format(
        index_name=get_index_name(viewdef.name, unique=True),
        view_name=viewdef.name,
        index_keys=index_keys,
    )

    return dedent(query)


def get_all_views_query() -> str:
    """ doc """
    # There is a better way to do this to get out deps
    return "select table_name from INFORMATION_SCHEMA.views where table_schema='public';"


def get_all_views_deps() -> str:
    """ doc """

    query = """
    with view_oids as (
        select
            distinct(dependent_view.oid) as view_oid

        from pg_depend
        JOIN pg_rewrite ON pg_depend.objid = pg_rewrite.oid
        JOIN pg_class as dependent_view ON pg_rewrite.ev_class = dependent_view.oid
        JOIN pg_namespace dependent_ns ON dependent_ns.oid = dependent_view.relnamespace
        WHERE
        dependent_ns.nspname = 'public'
    ), view_dependencies as (
        select
            dependent_view.oid as dependent_oid,
            dependent_ns.nspname as dependent_schema,
            dependent_view.relname as dependent_view,
            source_table.oid as dependency_oid,
            source_ns.nspname as source_schema,
            source_table.relname as source_view
        from pg_depend
        JOIN pg_rewrite ON pg_depend.objid = pg_rewrite.oid
        JOIN pg_class as dependent_view ON pg_rewrite.ev_class = dependent_view.oid
        JOIN pg_class as source_table ON pg_depend.refobjid = source_table.oid
        JOIN view_oids on source_table.oid = view_oids.view_oid
        JOIN pg_namespace dependent_ns ON dependent_ns.oid = dependent_view.relnamespace
        JOIN pg_namespace source_ns ON source_ns.oid = source_table.relnamespace
        WHERE
            source_ns.nspname = 'public'
        group by
            dependent_view.oid,
            dependent_ns.nspname,
            dependent_view.relname,
            source_table.oid,
            source_ns.nspname,
            source_table.relname
    )
    select
        view_dependencies.*
    from view_dependencies
    ;
    """

    return dedent(query)



"""
OpenNEM cache utilities

"""
import logging
from functools import wraps
from typing import Callable, Dict, List, Optional

from cachetools import TTLCache

from opennem.api.stats.schema import ScadaDateRange
from opennem.schema.network import NetworkSchema
from opennem.settings import settings

logger = logging.getLogger(__name__)

CACHE_AGE = settings.cache_scada_values_ttl_sec

scada_cache: TTLCache = TTLCache(maxsize=100, ttl=CACHE_AGE)


def cache_scada_result(func: Callable) -> Callable:
    """
    Caches the scada_range results in memory since they're called so often
    by wrapping the function.
    """

    @wraps(func)
    def _cache_scada_wrapper(
        network: Optional[NetworkSchema] = None,
        networks: Optional[List[NetworkSchema]] = None,
        network_region: Optional[str] = None,
        facilities: Optional[List[str]] = None,
    ) -> Optional[ScadaDateRange]:
        key_list = []

        if network:
            key_list = [network.code]

        if networks:
            key_list += [n.code for n in networks]

        if facilities:
            key_list += facilities

        key = frozenset(key_list)
        ret: Optional[ScadaDateRange] = None

        try:
            _val: Dict = scada_cache[key]
            ret = ScadaDateRange(**_val)

            logger.debug("scada range HIT at key: {}".format(key))

            return ret
        except KeyError:
            ret = func(network, networks, network_region, facilities)

            logger.debug("scada range MISS at key: {}".format(key))

            if ret:
                scada_cache[key] = ret.dict()
            return ret

    return _cache_scada_wrapper



import json
import logging
from urllib.parse import urlparse

import requests
import wikipedia
from PIL import Image

from opennem.api.photo.controllers import write_photo_to_s3
from opennem.core.loader import load_data
from opennem.core.normalizers import station_name_cleaner
from opennem.db import SessionLocal
from opennem.db.models.opennem import Photo, Station
from opennem.utils.images import img_to_buffer

logger = logging.getLogger(__name__)


def article_from_wikipedia(url: str) -> str:
    """
    Return the article name from a wikipedia url
    """
    path = urlparse(url).path

    if not path:
        return ""

    return path.split("/")[2]


def dataid_from_url(url: str) -> str:
    """
    Return the Q id from a wikidata url
    """
    path = urlparse(url).path

    if not path:
        return ""

    return path.split("/")[2]


def wikidata_join():
    session = SessionLocal()

    wikidata = load_data("wikidata-parsed.json", from_project=True)

    # session.add()

    for entry in wikidata:
        station_name = entry.get("name")

        station_lookup = (
            session.query(Station).filter(Station.name == station_name).all()
        )

        if len(station_lookup) == 0:
            logger.info("Didn't find a station for {}".format(station_name))

        if len(station_lookup) == 1:
            station = station_lookup.pop()

            station.description = entry.get("description")
            station.wikipedia_link = entry.get("wikipedia")
            station.wikidata_id = entry.get("wikidata_id")

            session.add(station)
            logger.info("Updated station {}".format(station_name))

        if len(station_lookup) > 1:
            logger.info("Found multiple for station {}".format(station_name))

    session.commit()


def wikidata_parse():

    # query: https://w.wiki/dVi
    wikidata = load_data("wikidata.json", from_project=True)

    out_entries = []
    total_entries = len(wikidata)
    current = 0

    for entry in wikidata:
        wikilink = article_from_wikipedia(entry["article"])
        wikidata = dataid_from_url(entry["item"])
        station_name = station_name_cleaner(entry["itemLabel"])

        description = None

        try:
            description = wikipedia.summary(wikilink)
        except Exception as e:
            print(e)

        new_entry = {
            "wikipedia": entry["article"],
            "wikidata": entry["item"],
            "wiki_id": wikilink,
            "wikidata_id": wikidata,
            "name": station_name,
            "name_original": entry["itemLabel"],
            "description": description,
        }

        out_entries.append(new_entry)
        current += 1

        print("Done {} of {}".format(current, total_entries))

    with open("data/wikidata-parsed.json", "w") as fh:
        json.dump(out_entries, fh)


def get_image(image_url):
    img = None

    try:
        img = Image.open(requests.get(image_url, stream=True).raw)
    except Exception:
        logger.error("Error parsing: %s", image_url)
        return None

    return img


def wikidata_photos():
    session = SessionLocal()
    wikidata = load_data("wikidata-photos.json", from_project=True)

    for entry in wikidata:
        image_url = entry["thumb"]
        name = entry["itemLabel"]
        wiki_id = dataid_from_url(entry["item"])

        station = (
            session.query(Station)
            .filter(Station.wikidata_id == wiki_id)
            .one_or_none()
        )

        if not station:
            print("Could not find station {}".format(name))
            # continue

        img = get_image(image_url)

        if not img:
            print("No image for {}".format(name))
            continue

        # file_name = urlparse(image_url).path.split("/")[-1:]
        file_name = "{}_{}.{}".format(
            name.replace(" ", "_"), "original", "jpeg"
        )

        photo = Photo(
            name=file_name,
            width=img.size[0],
            height=img.size[1],
            original_url=image_url,
        )

        img_buff = img_to_buffer(img)

        write_photo_to_s3(file_name, img_buff)

        if station:
            station.photos.append(photo)

        img.thumbnail((280, 340))

        file_name = "{}_{}.{}".format(
            name.replace(" ", "_"), img.size[0], "jpeg"
        )

        photo_thumb = Photo(
            name=file_name,
            width=img.size[0],
            height=img.size[1],
            original_url=image_url,
        )

        img_buff = img_to_buffer(img)
        write_photo_to_s3(file_name, img_buff)

        if station:
            station.photos.append(photo_thumb)

        session.add(photo)
        session.add(photo_thumb)

        if station:
            session.add(station)

        session.commit()


if __name__ == "__main__":
    wikidata_join()
    wikidata_photos()
